from typing import ClassVar, Literal, Self

from . import Color
from .style import Layer


class AnsiColor:
    """
    The 16 extended ANSI colors.

    The ANSI color comprise eight base colors and their eight bright versions.
    Unlike most other colors, the ANSI colors have no intrinsic color values and
    hence are abstract colors. At the same time, they do have semantics.
    """
    Black: ClassVar[AnsiColor] = ...
    Red: ClassVar[AnsiColor] = ...
    Green: ClassVar[AnsiColor] = ...
    Yellow: ClassVar[AnsiColor] = ...
    Blue: ClassVar[AnsiColor] = ...
    Magenta: ClassVar[AnsiColor] = ...
    Cyan: ClassVar[AnsiColor] = ...
    White: ClassVar[AnsiColor] = ...
    BrightBlack: ClassVar[AnsiColor] = ...
    BrightRed: ClassVar[AnsiColor] = ...
    BrightGreen: ClassVar[AnsiColor] = ...
    BrightYellow: ClassVar[AnsiColor] = ...
    BrightBlue: ClassVar[AnsiColor] = ...
    BrightMagenta: ClassVar[AnsiColor] = ...
    BrightCyan: ClassVar[AnsiColor] = ...
    BrightWhite: ClassVar[AnsiColor] = ...

    @staticmethod
    def all() -> AnsiColorIterator: ...
    @staticmethod
    def try_from_8bit(value: int) -> AnsiColor: ...
    def to_8bit(self) -> int: ...
    def is_achromatic(self) -> bool: ...
    def is_bright(self) -> bool: ...
    def to_3bit(self) -> AnsiColor: ...
    def name(self) -> str: ...
    def abbr(self) -> str: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __le__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...


class AnsiColorIterator:
    """An iterator over all ANSI colors."""
    def __len__(self) -> int: ...
    def __iter__(self) -> Self: ...
    def __next__(self) -> AnsiColor: ...


class EmbeddedRgb:
    """The 6x6x6 RGB cube embedded in 8-bit color."""
    def __new__(cls, r: int, g: int, b:int) -> Self: ...
    @staticmethod
    def try_from_8bit(value: int) -> EmbeddedRgb: ...
    def to_8bit(self) -> int: ...
    def to_24bit(self) -> bytes: ...
    def to_color(self) -> Color: ...
    def coordinates(self) -> bytes: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> int: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...


class GrayGradient:
    """The 24-step gray gradient embedded in 8-bit color."""
    def __new__(cls, value: int) -> Self: ...
    @staticmethod
    def try_from_8bit(value: int) -> GrayGradient: ...
    def to_8bit(self) -> int: ...
    def to_24bit(self) -> bytes: ...
    def to_color(self) -> Color: ...
    def level(self) -> int: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __le__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...


class EightBitColor_Ansi(EightBitColor):
    """The ANSI colors."""
    def __new__(cls, color: AnsiColor) -> Self: ...
    def __getitem__(self, index: Literal[0]) -> AnsiColor: ...


class EightBitColor_Embedded(EightBitColor):
    """The embedded RGB colors."""
    def __new__(cls, color: EmbeddedRgb) -> Self: ...
    def __getitem__(self, index: Literal[0]) -> EmbeddedRgb: ...


class EightBitColor_Gray(EightBitColor):
    """The gray gradient colors."""
    def __new__(cls, color: GrayGradient) -> Self: ...
    def __getitem__(self, index: Literal[0]) -> GrayGradient: ...


class EightBitColor:
    """The 8-bit colors."""
    Ansi = EightBitColor_Ansi
    Embedded = EightBitColor_Embedded
    Gray = EightBitColor_Gray

    @staticmethod
    def from_8bit(byte: int) -> EightBitColor: ...

    def to_8bit(self) -> int: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...


class Rgb:
    """24-bit RGB colors."""
    def __new__(cls, r: int, g: int, b: int) -> Rgb: ...
    @staticmethod
    def from_color(color: Color) -> Rgb: ...
    def to_color(self) -> Color: ...
    def coordinates(self) -> bytes: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> int: ...
    def weighted_euclidian_distance(self, other: Rgb) -> int: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...


class Colorant_Default(Colorant):
    """The default color."""
    def __new__(cls) -> Self: ...


class Colorant_Ansi(Colorant):
    """The ANSI colors."""
    def __new__(cls, color: AnsiColor) -> Self: ...
    def __getitem__(self, index: Literal[0]) -> AnsiColor: ...


class Colorant_Embedded(Colorant):
    """The embedded RGB colors."""
    def __new__(cls, color: EmbeddedRgb) -> Self: ...
    def __getitem__(self, index: Literal[0]) -> EmbeddedRgb: ...


class Colorant_Gray(Colorant):
    """The gray gradient colors."""
    def __new__(cls, color: GrayGradient) -> Self: ...
    def __getitem__(self, index: Literal[0]) -> GrayGradient: ...


class Colorant_Rgb(Colorant):
    """The 24bit colors."""
    def __new__(cls, color: Rgb) -> Self: ...
    def __getitem__(self, index: Literal[0]) -> Rgb: ...


class Colorant_HiRes(Colorant):
    """The high-resolution colors."""
    def __new__(cls, color: Color) -> Self: ...
    def __getitem__(self, index: Literal[0]) -> Color: ...


class Colorant:
    """A wrapper for all color representations."""
    Default = Colorant_Default
    Ansi = Colorant_Ansi
    Embedded = Colorant_Embedded
    Gray = Colorant_Gray
    Rgb = Colorant_Rgb
    HiRes = Colorant_HiRes

    @staticmethod
    def of(
        colorant: (
            int | AnsiColor | EmbeddedRgb | GrayGradient | EightBitColor | Rgb
            | Color | Colorant
        )
    ) -> Colorant: ...

    def try_to_8bit(self) -> int: ...
    def try_to_24bit(self) -> tuple[int, int, int]: ...
    def is_default(self) -> bool: ...
    def __neg__(self) -> None | Colorant: ...
    def __getitem__(
        self, index: Literal[0]
    ) -> AnsiColor | EmbeddedRgb | GrayGradient | Rgb | Color: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...
    def display(self, layer: Layer) -> str: ...



"""
A script to visualize 8-bit terminal colors as well as prettypretty's support
for down-sampling colors and maximizing contrast.
"""
import argparse
import enum
import math
from typing import cast, Literal

from .color import Color
from .color.style import (Fidelity, Layer) # pyright: ignore [reportMissingModuleSource]
from .color.termco import (AnsiColor, EmbeddedRgb, Rgb) # pyright: ignore [reportMissingModuleSource]
from .color.theme import VGA_COLORS # pyright: ignore [reportMissingModuleSource]
from .theme import current_translator
from .terminal import Terminal


class AnsiConversion(enum.Enum):
    """The strategy for converting embedded RGB to ANSI colors."""
    NoConversion = enum.auto()
    HueLightness = enum.auto()
    OklabDistance = enum.auto()
    RgbDistance = enum.auto()
    RgbRounding = enum.auto()


def show_error(term: Terminal, msg: str) -> None:
    """Show a big boxy red error message"""
    def line(s: str) -> None:
        (
            term
            .bold()
            .fg(EmbeddedRgb(5, 5, 5))
            .bg(AnsiColor.Red)
            .write(s)
            .reset_style()
            .writeln()
        )

    line(" " * (len(msg) + 4))
    line(f"  {msg}  ")
    line(" " * (len(msg) + 4))


class FramedBoxes:
    """
    Emit boxes in a frame.

    This class helps with incrementally emitting a grid of boxes with a
    surrounding frame (or border). Each box has the same width and a height of
    one line. Boxes are formatted left-to-right.
    """
    def __init__(
        self,
        term: Terminal,
        box_count: int = 1,
        min_width: int = 5,
        max_width: int = 1_000,
    ) -> None:
        self._term = term
        self._box_count = box_count
        self._box_width = min(term.width - 2, max_width) // box_count
        if self._box_width < min_width:
            raise ValueError(
                f'unable to fit {box_count} boxes into {term.width} columns'
            )
        self._width = self._box_count * self._box_width + 2
        self._indent = (term.width - self.outer_width) // 2
        self._line_content_width = 0

    @property
    def outer_width(self) -> int:
        """The total width of the frame."""
        return self._width

    @property
    def inner_width(self) -> int:
        """
        The width of the content inside the frame, which is ``outer_width - 2``.
        """
        return self._width - 2

    def top(self, title: str = '') -> None:
        """Format the top of the frame."""
        title = f' {t} ' if (t := title.strip()) != '' else t
        filling = self.outer_width - 4 - len(title)
        if filling < 0:
            raise ValueError(f'"{title}" is too long for {self.outer_width}-wide frame')
        if title:
            title = f'\x1b[1m{title}\x1b[m'
        self._term.writeln(f'{" " * self._indent}┏━{title}{"━" * filling}━┓')

    def left(self) -> None:
        """Start formatting a line of content."""
        if self._line_content_width != 0:
            raise ValueError('Line started before line ended')
        self._term.write(f'{" " * self._indent}┃')

    def box(
        self,
        text: str,
        foreground: tuple[int] | tuple[int, int, int],
        background: tuple[int] | tuple[int, int, int]
    ) -> None:
        """Format one box of the content."""
        box = text.center(self._box_width)
        if len(box) != self._box_width:
            raise ValueError(f'"{text}" does not fit into {self._box_width}-wide box')

        self._term.fg(*foreground).bg(*background).bold().write(box)
        self._line_content_width += self._box_width

    def right(self) -> None:
        """Complete formatting a line of content."""
        self._term.reset_style()
        self._term.writeln('┃')

        if self._line_content_width != self.inner_width:
            raise ValueError(
                f'content spans {self._line_content_width}, '
                f'not {self.inner_width} columns'
            )

        self._line_content_width = 0

    def bottom(self) -> None:
        """Format the bottom of the frame."""
        self._term.writeln(f'{" " * self._indent}┗{"━" * self.inner_width}┛')


def write_color_cube(
    term: Terminal,
    *,
    layer: Layer = Layer.Background,
    strategy: AnsiConversion = AnsiConversion.NoConversion,
    show_detail: Literal[0, 1, 2] = 2,
) -> None:
    """
    Format a framed grid with 216 cells, where each cell displays a distinct
    color from the 6x6x6 cube of 8-bit terminal colors.

    Args:
        term: is the terminal for write the framed grid to
        layer: determines whether to color text or background, with background
            the default
        strategy: determines whether to display the original 8-bit colors, to
            downsample using prettypretty, or to use the naive RGB conversion
        label: determines whether boxes are labelled with their color
            components
    """
    translator = current_translator()
    if (
        not translator.supports_hue_lightness()
        and strategy is AnsiConversion.HueLightness
    ):
        show_error(
            term,
            "Terminal theme violates requirements of hue/lightness algorithm. "
            "Skipping frame."
        )
        return

    theme_colors: list[Rgb] = [
        Rgb.from_color(translator.resolve(i)) for i in range(16)
    ]

    def closest_theme_color(color: Rgb) -> int:
        min_distance = math.inf
        min_index = -1

        for index, theme_color in enumerate(theme_colors):
            distance = color.weighted_euclidian_distance(theme_color)
            if distance < min_distance:
                min_distance = distance
                min_index = index

        return min_index

    min_width = 5 if show_detail > 0 else 8
    max_width = 1_000 if show_detail > 0 else 48
    frame = FramedBoxes(term, box_count=6, min_width=min_width, max_width=max_width)

    title = (
        'Original' if strategy == AnsiConversion.NoConversion
        else 'Hue/Lightness' if strategy == AnsiConversion.HueLightness
        else 'Oklab Distance' if strategy == AnsiConversion.OklabDistance
        else 'RGB Distance' if strategy == AnsiConversion.RgbDistance
        else 'Round to 1-bit'
    )

    if show_detail > 0:
        title = f'{layer}: 6•6•6 RGB Cube ({title})'
    frame.top(title)

    for r in range(6):
        for b in range(6):
            for _ in range(1 if show_detail > 0 else 4):
                frame.left()

                for g in range(6):
                    embedded = EmbeddedRgb(r, g, b)
                    source = embedded.to_color()

                    if strategy == AnsiConversion.NoConversion:
                        eight_bit = embedded.to_8bit()
                    elif strategy == AnsiConversion.HueLightness:
                        maybe_value = translator.to_ansi_hue_lightness(source)
                        if maybe_value is None:
                            raise ValueError('hue/lightness reduction unavailable')
                        eight_bit = maybe_value.to_8bit()
                    elif strategy == AnsiConversion.OklabDistance:
                        eight_bit = translator.to_closest_ansi(source).to_8bit()
                    elif strategy == AnsiConversion.RgbDistance:
                        eight_bit = closest_theme_color(Rgb(*embedded.to_24bit()))
                    elif strategy == AnsiConversion.RgbRounding:
                        eight_bit = translator.to_ansi_rgb(source).to_8bit()
                    else:
                        raise ValueError(f'invalid strategy "{strategy}"')

                    target = translator.resolve(eight_bit)

                    # Pick black or white for target, not source color.
                    if layer is Layer.Background:
                        foreground = 16 if target.use_black_text() else 231,
                        background = eight_bit,
                    else:
                        foreground = eight_bit,
                        background = 16 if target.use_black_background() else 231,

                    frame.box(
                        f'{r}•{g}•{b}' if show_detail == 2 else ' ',
                        foreground,
                        background,
                    )

                frame.right()

    frame.bottom()
    term.writeln()


def write_hires_slice(
    term: Terminal,
    *,
    hold: Literal['r', 'g', 'b'] = 'g',
    level: int = 0,
    eight_bit_only: bool = False,
) -> None:
    frame = FramedBoxes(term, 32, min_width=1)
    label = '/'.join(
        f'{axis.upper()}={level}' if axis == hold else axis.upper()
        for axis in ('r', 'g', 'b')
    )
    frame.top(
        ('Downsampled ' if eight_bit_only else '')
        + 'Hi-Res Color Slice for '
        + label
    )

    translator = current_translator()

    def emit_box(r: int, g: int, b: int) -> None:
        if eight_bit_only:
            color = translator.to_closest_8bit(Color.from_24bit(r, g, b)).to_8bit(),
        else:
            color = r, g, b

        frame.box(' ', (0,), color)

    for x in range(0, 256, 8):
        frame.left()
        for y in range(0, 256, 8):
            if hold == 'r':
                r = level
                g = x
                b = y
            elif hold == 'g':
                r = x
                g = level
                b = y
            elif hold == 'b':
                r = x
                g = y
                b = level
            else:
                raise AssertionError(f'invalid hold "{hold}"')

            emit_box(r, g, b)

        frame.right()

    frame.bottom()
    term.writeln()


def write_theme_test(term: Terminal, show_detail: Literal[0, 1, 2] = 2):
    frame = FramedBoxes(term, 2, max_width=40)
    frame.top('Actual vs Claimed Color')

    translator = current_translator()

    for index in range(16):
        color = translator.resolve(index)
        fg = 16 if color.use_black_text() else 231
        bg = color.to_24bit()

        label = ', '.join(f'{c:3d}' for c in bg)

        frame.left()
        frame.box(f'{index:>2}' if show_detail == 2 else ' ', (fg,), (index,))
        frame.box(
            label if show_detail == 2 else ' ',
            (fg,),
            cast(tuple[int, int, int], bg)
        )
        frame.right()

    frame.bottom()
    term.writeln()

def create_parser() -> argparse.ArgumentParser:
    """Create a command line argument parser."""
    parser = argparse.ArgumentParser(
        description="""
            Display color grids that visualize the range of terminal colors,
            while also exercising prettypretty's support for maximizing contrast
            and down-sampling colors. By default, this script uses the
            terminal's current color theme and makes a best guess as to the
            terminal's level of color support. That guess is likely less than
            accurate when running ssh.
        """,
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--vga',
        action='store_const',
        const=VGA_COLORS,
        dest='theme',
        help='use the same colors as VGA in text mode'
    )

    parser.add_argument(
        '--truecolor',
        action=argparse.BooleanOptionalAction,
        default=None,
        help='ignore advertised capabilities and force/suppress 24-bit mode',
    )
    parser.add_argument(
        '--slices',
        action='store_true',
        help='show slices through 24-bit RGB cube (in truecolor mode only)'
    )
    parser.add_argument(
        '--abstract',
        action='store_const',
        const=0,
        dest='detail',
        default=2,
        help='show smaller, label-free color grids',
    )
    parser.add_argument(
        '--no-label',
        action='store_const',
        const=1,
        dest='detail',
        help='do not display color labels',
    )

    return parser


if __name__ == '__main__':
    options = create_parser().parse_args()

    fidelity = None
    if options.truecolor is not None:
        fidelity = Fidelity.TwentyFourBit if options.truecolor else Fidelity.EightBit

    with (
        Terminal(fidelity=fidelity)
        .terminal_theme(options.theme)
        .scoped_style()
        as term
    ):
        term.writeln()

        write_color_cube(term, show_detail=options.detail)
        write_color_cube(term, strategy=AnsiConversion.HueLightness, show_detail=options.detail)
        write_color_cube(term, strategy=AnsiConversion.OklabDistance, show_detail=options.detail)
        write_color_cube(term, strategy=AnsiConversion.RgbDistance, show_detail=options.detail)
        write_color_cube(term, strategy=AnsiConversion.RgbRounding, show_detail=options.detail)
        write_color_cube(term, layer=Layer.Foreground)

        if term.fidelity == Fidelity.TwentyFourBit:
            if options.slices:
                for hold in ('r', 'g', 'b'):
                    for level in (0, 128, 255):
                        for downsample in (False, True):
                            write_hires_slice(
                                term,
                                hold=hold,
                                level=level,
                                eight_bit_only=downsample,
                            )

            write_theme_test(term, show_detail=options.detail)

            term.write_paragraph("""
                The frame showing actual vs claimed colors has two columns. The
                background colors of the first are the 16 extended ANSI colors.
                The background colors of the second are the corresponding 24-bit
                RGB colors advertised by the terminal.

                Both columns should have the exact same visual background color.
            """)

        color_mode = (
            'truecolor' if term.fidelity is Fidelity.TwentyFourBit else '8-bit color'
        )
        term.writeln('The above charts utilize ', color_mode, ' mode!\n')

        color_support = None
        with term.cbreak_mode():
            color_support = term.request_color_support()
        if color_support is None:
            support_level = 'does not respond to style queries'
        else:
            support_level = f'reports support for {color_support}'

        term.writeln(f'The terminal {support_level}!').flush()



import os
import subprocess
import sys


def identify_terminal() -> None | tuple[str, str]:
    """
    Identify the current terminal.

    This function implements the fallback strategies for
    :meth:`.Terminal.request_terminal_identity`.
    """
    name = lookup_term_program()
    version = lookup_term_program_version()
    if name and version:
        return normalize_terminal_name(name), version

    if sys.platform != "darwin":
        return None

    bundle = lookup_macos_bundle_id()
    if bundle is None:
        return None

    version = lookup_macos_bundle_version(bundle)
    return normalize_terminal_name(bundle), version or ''


def _init_registry() -> dict[str, str]:
    registry: dict[str, str] = {}

    for name, *aliases in (
        ("Alacritty", "org.alacritty", "org.alacritty.Alacritty"),
        ("Hyper", "co.zeit.hyper"),
        ("iTerm", "com.googlecode.iterm2", "iTerm2"),
        ("Kitty", "net.kovidgoyal.kitty"),
        ("Rio", "com.raphaelamorim.rio"),
        ("Tabby", "org.tabby"),
        ("Terminal.app", "com.apple.Terminal", "Apple_Terminal"),
        ("VSCode", "com.microsoft.VSCode", "Code", "Visual Studio Code"),
        ("Warp", "dev.warp.Warp-Stable", "WarpTerminal"),
        ("WezTerm", "com.github.wez.wezterm", "org.wezfurlong.wezterm"),
    ):
        registry[name.casefold()] = name
        for alias in aliases:
            registry[alias.casefold()] = name

    return registry

_REGISTRY = _init_registry()


def normalize_terminal_name(name: str) -> str:
    """
    Normalize the terminal name or bundle ID.

    If the given name is an alias for a well-known terminal, this function
    returns the canonical name. Otherwise, it just returns the given name.
    """
    normal = _REGISTRY.get(name.casefold())
    return normal if normal else name


def lookup_term_program() -> None | str:
    """Look up the term program."""
    return os.getenv('TERM_PROGRAM')


def lookup_term_program_version() -> None | str:
    """Look up the term program version."""
    return os.getenv('TERM_PROGRAM_VERSION')


def lookup_macos_bundle_id() -> None | str:
    """Look up the macOS bundle identifier for the current terminal."""
    return os.getenv('__CFBundleIdentifier')


def lookup_macos_bundle_version(bundle: str) -> None | str:
    """
    Look up the macOS bundle version for the given bundle ID.

    This function only runs on macOS.
    """
    if sys.platform != "darwin":
        raise NotImplementedError("only runs on macOS")

    # Locate actual bundle...
    paths = subprocess.run(
        ["mdfind", f"kMDItemCFBundleIdentifier == '{bundle}'"],
        stdout=subprocess.PIPE,
        encoding="utf8",
    ).stdout.strip().splitlines()

    if not paths:
        return None

    # ...to extract its version
    for path in paths:
        version = subprocess.run(
            ["mdls", "-name", "kMDItemVersion", "-raw", path],
            stdout=subprocess.PIPE,
            encoding="utf8",
        ).stdout.strip()

        if version:
            return version

    return None



from typing import Self

from . import Color
from .style import AnsiColor, Layer


class ThemeEntry_DefaultForeground(ThemeEntry):
    """A theme entry for one of the default colors."""
    def __new__(cls) -> Self: ...


class ThemeEntry_DefaultBackground(ThemeEntry):
    """A theme entry for one of the default colors."""
    def __new__(cls) -> Self: ...


class ThemeEntry_Ansi(ThemeEntry):
    """A theme entry for one of the ANSI colors."""
    def __new__(cls, color: AnsiColor) -> Self: ...


class ThemeEntry:
    """The 18 colors in a theme."""
    DefaultForeground = ThemeEntry_DefaultForeground
    DefaultBackground = ThemeEntry_DefaultBackground
    Ansi = ThemeEntry_Ansi

    @staticmethod
    def all() -> ThemeEntryIterator: ...
    @staticmethod
    def try_from_index(index: int) -> ThemeEntry: ...
    def name(self) -> str: ...
    def abbr(self) -> str: ...
    def parse(self, s: bytes) -> Color: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __le__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...


class ThemeEntryIterator:
    """An iterator over all theme entries."""
    def __len__(self) -> int: ...
    def __iter__(self) -> Self: ...
    def __next__(self) -> ThemeEntry: ...


class Theme:
    """A color theme."""
    @staticmethod
    def query_terminal() -> Theme: ...

    def __new__(cls, colors: list[Color]) -> Self: ...
    def __getitem__(self, index: ThemeEntry|AnsiColor|Layer) -> Color: ...
    def __setitem__(self, index: ThemeEntry|AnsiColor|Layer, color: Color) -> Color: ...
    def __repr__(self) -> str: ...


VGA_COLORS: Theme =  ...



from collections.abc import Iterator, Sequence
from contextlib import AbstractContextManager, contextmanager, ExitStack
import enum
import os
import select
import sys
import termios
import textwrap
import tty
from typing import (
    Any,
    Callable,
    cast,
    ClassVar,
    ContextManager,
    Literal,
    Never,
    overload,
    Self,
    TextIO,
    TypeAlias,
)

from .ansi import Ansi, RawAnsi
from .color import Color, theme # pyright: ignore [reportMissingModuleSource]
from .color.style import (Fidelity, Layer, Style) # pyright: ignore [reportMissingModuleSource]
from .color.termco import ( # pyright: ignore [reportMissingModuleSource]
    AnsiColor, Colorant, EmbeddedRgb, GrayGradient, Rgb
)
from .theme import new_theme, current_translator
from .ident import identify_terminal, normalize_terminal_name


TerminalMode: TypeAlias = list[Any]


class TerminalModeComponent:
    CC = 6
    LFLAG = 3


_REFLECTED_METHODS = {
    'write_control', 'at', 'column', 'down', 'left', 'link', 'right', 'up'
}


class BatchMode(enum.Enum):
    """
    A terminal's `batch mode
    <https://gist.github.com/christianparpart/d8a62cc1ab659194337d73e399004036>`_.

    Attributes:
        NOT_SUPPORTED: indicates that the terminal does not support batching
        ENABLED: indicates that the terminal is currently batching
        DISABLED: indicates that the terminal is not currently batching
        UNDEFINED: really is permanently enabled, which makes no sense
        PERMANENTLY_DISABLED: effectively is the same as the terminal not
            supporting batch mode

    Since three out of five status codes for batching more (``NOT_SUPPORTED``,
    ``PERMANENTLY_DISABLED``) or less (``UNDEFINED``) imply that the terminal
    doesn't support batching, this enumeration also defines the more precise
    computed properties ``is_supported``, ``is_enabled``, and ``is_disabled``.
    """
    NOT_SUPPORTED = 0
    ENABLED = 1
    DISABLED = 2
    UNDEFINED = 3
    PERMANENTLY_DISABLED = 4

    @property
    def is_supported(self) -> bool:
        """Determine whether the terminal supports batching."""
        return self is BatchMode.ENABLED or self is BatchMode.DISABLED

    @property
    def is_enabled(self) -> bool:
        """Determine whether the terminal is currently batching."""
        return self is BatchMode.ENABLED

    @property
    def is_disabled(self) -> bool:
        """Determine whether the terminal is currently not batching."""
        return self is not BatchMode.ENABLED


class TerminalContextManager(AbstractContextManager['Terminal']):
    """
    A context manager for terminal state.

    This class manages operations that update and restore terminal state. It
    ensures that all updates are applied on entry and restored on exit, that
    updates are applied in registration order and restored in the opposite
    order, and that terminal output is flushed after applying and also after
    reverting all updates.

    An update may be a pair of escape sequences or a function that instantiates
    a context manager. In case of the former, the first sequence is written to
    the output upon entry and the second sequence upon exit. In case of the
    latter, the function and its result's ``__enter__()`` method are invoked on
    entry and the result's ``__exit__()`` method is invoked on exit. Currently,
    there is no public interface for registering the latter.

    This class is reentrant and reusable, though an instance does nothing on
    nested invocations.

    :class:`Terminal` has several methods with the same names and signatures as
    this class. They are the preferred way of creating terminal context manager
    instances because they are far more convenient. For example, to get started
    using prettypretty, you might write:

    .. code-block:: python

        terminal = Terminal()
        with TerminalContextManager(terminal).terminal_theme().scoped_style():
            ...

    is equivalent to this far nicer alternative

    .. code-block:: python

        with Terminal().alternate_screen().hidden_cursor() as terminal:
            ...

    """
    def __init__(self, terminal: 'Terminal') -> None:
        self._terminal = terminal
        self._updates: list[Callable[[], ContextManager[object]] | tuple[str, str]] = []
        self._block_depth = 0
        self._exit_stack: None | ExitStack = None

    def _check_not_active(self) -> None:
        if self._block_depth > 0:
            raise ValueError(
                'unable to update context manager after __enter__() has been called'
            )

    def register(self, do: str, undo: str) -> Self:
        """
        Register an update with this terminal context manager. Both ``do`` and
        ``undo`` should be ANSI escape sequences, with ``undo`` restoring the
        terminal to the state from before ``do``.
        """
        self._check_not_active()
        self._updates.append((do, undo))
        return self

    @contextmanager
    def _cbreak_mode(self) -> 'Iterator[Terminal]':
        fileno = self._terminal._input_fileno  # type: ignore[reportPrivateUsage]
        saved_mode = termios.tcgetattr(fileno)
        change_mode = not self._terminal.is_cbreak_mode(saved_mode)
        if change_mode:
            tty.setcbreak(fileno)
        try:
            yield self._terminal
        finally:
            if change_mode:
                termios.tcsetattr(fileno, termios.TCSAFLUSH, saved_mode)

    def cbreak_mode(self) -> Self:
        """
        Put the terminal into cbreak mode.

        If the terminal is not yet in cbreak mode, the context manager sets
        cbreak mode upon entry and restores the previous mode upon exit. If the
        terminal is in cbreak mode already, the context manager does not modify
        the terminal mode. Mode changes only take effect after all queued output
        has been written but queued input is discarded.
        """
        self._check_not_active()
        self._updates.append(lambda: self._cbreak_mode())
        return self

    def _request_theme(self) -> theme.Theme:
        with self._cbreak_mode():
            return self._terminal.request_theme()

    def terminal_theme(self, theme: None | theme.Theme = None) -> Self:
        """
        Use the terminal's color theme. Unless a theme is provided as argument,
        the context manager puts the terminal temporarily into cbreak mode and
        determines the current theme colors upon entry. It then makes that theme
        the current theme until exit.
        """
        self._check_not_active()
        if theme is None:
            factory = lambda: new_theme(self._request_theme())
        else:
            factory = lambda: new_theme(theme)
        self._updates.append(factory)
        return self

    def window_title(self, title: str) -> Self:
        """Update the window title."""
        # Save window title on stack, then update window title
        return self.register(
            Ansi.fuse(Ansi.CSI, "22;2t", Ansi.OSC, "0;", title, Ansi.ST),
            Ansi.fuse(Ansi.CSI, "23;2t"),
        )

    def alternate_screen(self) -> Self:
        """Switch to the terminal's alternate (unbuffered) screen."""
        return self.register(
            Ansi.fuse(Ansi.CSI, "?1049h"),
            Ansi.fuse(Ansi.CSI, "?1049l"),
        )

    def hidden_cursor(self) -> Self:
        """Make cursor invisible."""
        return self.register(
            Ansi.fuse(Ansi.CSI, "?25l"),
            Ansi.fuse(Ansi.CSI, "?25h"),
        )

    def batched_output(self) -> Self:
        """
        Batch terminal output.

        While batching, a terminal temporarily delays updating the screen by
        buffering output. It avoids visual artifacts when rapidly updating the
        screen.
        """
        return self.register(
            Ansi.fuse(Ansi.CSI, "?2026h"),
            Ansi.fuse(Ansi.CSI, "?2026l"),
        )

    def bracketed_paste(self) -> Self:
        """
        Enable `bracketed pasting
        <https://gitlab.com/gnachman/iterm2/-/wikis/Paste-Bracketing>`_.
        """
        return self.register(
            Ansi.fuse(Ansi.CSI, "?2004h"),
            Ansi.fuse(Ansi.CSI, "?2004l"),
        )

    def scoped_style(self) -> Self:
        """Scope style changes by resetting the style on exit."""
        return self.register('', Ansi.fuse(Ansi.CSI, 'm'))

    def _create_control_writer(self, control: str) -> Callable[[], None]:
        def control_writer() -> None:
            self._terminal.write_control(control)
        return control_writer

    def __enter__(self) -> 'Terminal':
        if self._block_depth == 0:
            with ExitStack() as stack:
                # Protect against partial terminal updates with an exit stack.
                for update in self._updates:
                    if isinstance(update, tuple):
                        self._terminal.write_control(update[0])
                        stack.callback(self._create_control_writer(update[1]))
                    else:
                        stack.enter_context(update())

                # All updates have been processed. Flush terminal output.
                self._terminal.flush()

                # All went well. Delay clean up until __exit__().
                self._exit_stack = stack.pop_all()

        self._block_depth += 1
        return self._terminal

    def __exit__(self, *args: Any) -> bool:
        # Exception suppression isn't needed right now but easy enough to support
        received_exception = args[0] is not None
        suppressed_exception = False

        if self._block_depth == 1:
            assert self._exit_stack is not None
            suppressed_exception = bool(self._exit_stack.__exit__(*args))
            self._terminal.flush()
            self._exit_stack = None

        self._block_depth -= 1
        return received_exception and suppressed_exception


class Terminal:
    """
    Terminal input/output.

    This class manages terminal input and output. It encapsulates the nitty
    gritty of managing the terminal with ANSI escape codes behind methods with
    meaningful, human-readable names instead of cryptic mnemonics (what does ED
    do again and how does it differ from DECSED?). Furthermore, many operations
    that would leave the terminal in an unusable state upon

    This class supports the following features:

    **Terminal window size**
        ``Terminal`` caches the width and height of the terminal. It only
        updates the cached values if an application explicitly polls for
        changes. That way, the application is hopefully prepared to accommodate
        a terminal size change as well.

        See :attr:`width`, :attr:`height`, :meth:`request_size`,
        :meth:`update_size`, and :meth:`check_same_size`.

    **Cbreak mode**
        In this mode, the terminal does not support line editing; instead, it
        immediately forwards bytes. As such cbreak mode facilitates
        request/response interactions between application and terminal. Getting
        authoritative information from the terminal sure beats other, more
        indirect ways, such as environment variables, that help surmise
        specific conditions.

        See :meth:`is_cbreak_mode`, :meth:`check_cbreak_mode`, and
        :meth:`cbreak_mode`.

    **Writing to terminal output**
        This class exposes different methods for writing text and for writing
        control sequences. The latter automatically fuses fragments together,
        adding semicolons between empty and numeric parameters. It also checks
        that the terminal supports ANSI escapes. Finally, it provides a
        convenient hook for intercepting them.

        See :meth:`write`, :meth:`writeln`:, :meth:`write_control`, and
        :meth:`flush`; also :attr:`fidelity`, :meth:`check_output_tty`, and
        :meth:`check_tty`.

    **Reading terminal input and ingesting ANSI escapes**
        Python's standard library has extensive support for reading from
        streams, but only blocking calls including for line-oriented input are
        convenient to use. This class makes up for that by implementing support
        for character-oriented, non-blocking input as well as for ANSI escape
        sequences. The latter require three levels of parsing:

         1. Parse individual character to read an entire control sequence, no
            less, no more.
         2. Parse message to to separate the integral or textual payload from
            message header and tail.
         3. Parse text to extract terminal name, colors, etc.

        See :meth:`read` and :meth:`read_control`; also
        :meth:`make_raw_request`, :meth:`parse_textual_response`, and
        :meth:`parse_numeric_response`; also :meth:`request_terminal_identity`,
        :meth:`request_cursor_position`, :meth:`request_batch_mode`,
        :meth:`request_ansi_color`, :meth:`request_dynamic_color`, and
        :meth:`request_theme`.

    **Scoped changes of terminal state**
        To more easily update, restore, and flush terminal states, ``Terminal``
        delegates to a terminal context manager. It makes it possible to
        fluently queue up all restorable updates in a single ``with`` statement
        without worrying about many of the implementation details.

        See :meth:`cbreak_mode`, :meth:`terminal_theme`, :meth:`window_title`,
        :meth:`alternate_screen`, :meth:`hidden_cursor`, :meth:`batched_output`,
        :meth:`bracketed_paste`, and :meth:`scoped_style`.

    **Simple updates of terminal state**
        Some terminal updates, notably for positioning the cursor and for
        erasing (parts of) the screen need not or can not be easily undone but
        still are eminently useful. You can also move the cursor and write links
        in :class:`.rich` text.

        See :meth:`up`, :meth:`down`, :meth:`left`, :meth:`right`,
        :meth:`set_position`, :meth:`set_column`, :meth:`erase_screen`,
        :meth:`erase_line`, and :meth:`link`.

    **Setting terminal styles**
        What's the point of integrating terminal colors with robust color
        management? Styling terminal output, of course! This class has methods
        to set bold, italic, or plain text and to set the fore/background
        colors. Those methods do not, however, adjust to the runtime context.
        For that, you want to use prettypretty's :class:`.Style` objects and
        :func:`rich` text.

        See :meth:`reset_style`, :meth:`rich_text`, :meth:`bold`,
        :meth:`italic`, :meth:`fg`, and :meth:`bg`.

    """
    def __init__(
        self,
        input: None | TextIO = None,
        output: None | TextIO = None,
        fidelity: None | Fidelity = None,
    ) -> None:
        self._identity: Literal[False] | None | tuple[str, str] = False

        self._input = cast(TextIO, input or sys.__stdin__)
        self._input_fileno = self._input.fileno()
        self._output = cast(TextIO, output or sys.__stdout__)
        self._all_tty = self._input.isatty() and self._output.isatty()

        if fidelity is None:
            fidelity = Fidelity.from_environment(self._output.isatty())
        self._fidelity = fidelity

        self._width, self._height = self.request_size() or (80, 24)

    # ----------------------------------------------------------------------------------

    @property
    def fidelity(self) -> Fidelity:
        """This terminal's color fidelity."""
        return self._fidelity

    def check_output_tty(self) -> Self:
        """Check that the output is a TTY."""
        if self._fidelity is Fidelity.Plain:
            raise ValueError('output does not accept ANSI escape sequences')
        return self

    def check_tty(self) -> Self:
        """Check that both input and output are TTYs."""
        if not self._all_tty:
            raise ValueError('input and output are not both TTYs')
        return self

    # ----------------------------------------------------------------------------------
    # Terminal Size

    @property
    def width(self) -> int:
        """The cached terminal width."""
        return self._width

    @property
    def height(self) -> int:
        """The cached terminal height."""
        return self._height

    def request_size(self) -> None | tuple[int, int]:
        """
        Determine the terminal's size in fixed-width columns and rows. If the
        underlying platform hook fails for both input and output, typically
        because both input and output have been redirected, this method returns
        ``None``.
        """
        try:
            return os.get_terminal_size(self._input_fileno)
        except OSError:
            pass
        try:
            return os.get_terminal_size(self._output.fileno())
        except OSError:
            return None

    def update_size(self) -> Self:
        """Update the width and height cached by this class."""
        self._width, self._height = self.request_size() or (80, 24)
        return self

    def check_same_size(self) -> Self:
        """
        Check that the terminal size has *not* changed since the last update.
        """
        w, h = self.request_size() or (80, 24)
        if self._width != w or self._height != h:
            raise AssertionError(
                f'terminal size changed from {self._width}×{self._height} to {w}×{h}'
            )
        return self

    # ----------------------------------------------------------------------------------

    def is_cbreak_mode(self, mode: None | TerminalMode = None) -> bool:
        """"
        Determine whether *cbreak mode* is enabled. This method inspects the
        current terminal mode to see whether characters are not echoed (``ECHO``
        is not set), line editing is disabled (``ICANON`` is not set), and reads
        return upon the first available character (``VMIN`` is 1, ``VTIME`` is
        0).

        Since raw mode makes the same changes and then some, this method detects
        raw mode as cbreak mode. That's just fine for its intended purpose,
        which is checking whether the terminal is prepared for handling ANSI
        escape sequences that require ANSI escape sequences as responses.
        """
        if mode is None:
            mode = termios.tcgetattr(self._input_fileno)

        # tty.setcbreak disables ECHO and ICANON, sets VMIN to 1, and VTIME to
        # 0. tty.setraw does the same and then some. For our purposes, it
        # suffices to check for cbreak's minimal settings, since we want the
        # terminal to respond right away and not stuck in line-buffer mode.
        return (
            not (mode[TerminalModeComponent.LFLAG] & termios.ECHO)
            and not (mode[TerminalModeComponent.LFLAG] & termios.ICANON)
            and mode[TerminalModeComponent.CC][termios.VMIN] == 1
            and mode[TerminalModeComponent.CC][termios.VTIME] == 0
        )

    def check_cbreak_mode(self) -> Self:
        """
        Check that cbreak mode is enabled. THis method signals an exception if
        cbreak mode is not enabled.
        """
        if not self.is_cbreak_mode():
            raise ValueError('terminal is expected to be in cbreak mode but is not')
        return self

    def cbreak_mode(self) -> TerminalContextManager:
        """
        Put the terminal into cbreak mode.

        If the terminal is not yet in cbreak mode, the context manager sets
        cbreak mode upon entry and restores the previous mode upon exit. If the
        terminal is in cbreak mode already, the context manager does not modify
        the terminal mode, but it still restores the previous mode upon exit.
        Mode changes only take effect after all queued output has been written
        but queued input is discarded.
        """
        return TerminalContextManager(self).cbreak_mode()

    # ----------------------------------------------------------------------------------

    def write(self, *fragments: str) -> Self:
        """
        Write the string fragments to this terminal's output. This method does
        not flush the output.
        """
        self._output.write(''.join(fragments))
        return self

    def writeln(self, *fragments: str) -> Self:
        """
        Write the string fragments followed by a line terminator to this
        terminal's output. This method does not flush the output.
        """
        self.write(*fragments, '\n')
        return self

    def write_paragraph(self, text: str) -> Self:
        """
        Write the paragraph to this terminal's output.

        This method strips all leading and trailing white space from each line
        of the text. It then treats each span of consecutive, non-empty lines as
        a paragraph and rewraps it to fit into the terminal width while still
        being convenient to read. Finally, it writes the resulting text to this
        terminal's output. This method does not flush the output.
        """
        lines = [line.strip() for line in text.splitlines()]
        start = stop = 0

        while True:
            # Skip empty lines
            while start < len(lines):
                if lines[start]:
                    break
                start += 1
            else:
                break

            # Determine span of non-empty lines
            stop = start + 1
            while stop < len(lines):
                if not lines[stop]:
                    break
                stop += 1

            # And rewrap the text in that span
            for line in textwrap.wrap(
                ' '.join(lines[start : stop]),
                width=min(self._width, 76),
                expand_tabs=False,
            ):
                self.writeln(line)

            # Repeat
            self.writeln()
            start = stop

        return self

    def write_control(self, *fragments: None | int | str) -> Self:
        """
        Write a control sequence to this terminal.

        This method :meth:`fuses <.Ansi.fuse>` the fragments of the inline
        control sequence (i.e., ANSI escape sequence) into a string and writes
        that string to this terminal's output. This method does not flush the
        terminal's output.

        This terminal's fidelity must not be :data:`Fidelity.PLAIN`, which is
        the case if the output is not a TTY. That restriction applies to all
        methods that write control sequences, since they always delegate to
        this method.
        """
        self.check_output_tty()
        self._output.write(Ansi.fuse(*fragments))
        return self

    def flush(self) -> Self:
        """Flush this terminal's output."""
        self._output.flush()
        return self

    # ----------------------------------------------------------------------------------

    def read(self, *, length: int = 3, timeout: float = 0) -> bytes:
        """
        Read raw bytes from this terminal.

        This method reads up to ``length`` bytes from this terminal. If the
        ``timeout`` is 0, this method does *not* wait for input and immediately
        returns, possibly with an empty byte string. If the ``timeout`` is
        greater than 0, this method does wait for input, up to as many seconds,
        using ``select()``.

        This terminal must be in cbreak mode.
        """
        self.check_cbreak_mode()
        if timeout > 0:
            ready, _, _ = select.select([self._input_fileno], [], [], timeout)
            if not ready:
                raise TimeoutError("timed out waiting to read from terminal input")
        return os.read(self._input_fileno, length)

    ESCAPE_TIMEOUT: ClassVar[float] = 0.5

    def read_control(self) -> bytes:
        """
        Read a complete ANSI escape sequence from this terminal.

        This method implements a reasonable but not entirely complete state
        machine for parsing ANSI escape sequences and keeps calling ``read()``
        for more bytes as necessary. It uses ``ESCAPE_TIMEOUT`` as timeout.

        The terminal must have TTYs for input and output. It also must be in
        cbreak mode.
        """
        self.check_tty().check_cbreak_mode()
        buffer = bytearray()

        def next_byte() -> int:
            b = self.read(length=1, timeout=self.ESCAPE_TIMEOUT)[0]
            buffer.append(b)
            return b

        def bad_byte(b: int) -> Never:
            raise ValueError(f"unexpected key code 0x{b:02X}")

        # TODO: Support ESC, CAN, SUB for cancellation

        b = next_byte()
        if b != 0x1B:
            bad_byte(b)

        # CSI Control Sequence
        # --------------------

        b = next_byte()
        if b == 0x5B:  # [
            b = next_byte()
            while 0x30 <= b <= 0x3F:
                b = next_byte()
            while 0x20 <= b <= 0x2F:
                b = next_byte()
            if 0x40 <= b <= 0x7E:
                return bytes(buffer)
            bad_byte(b)

        # DCS/SOS/OSC/PM/APC Control Sequence (Ending in ST)
        # --------------------------------------------------

        if b in (0x50, 0x58, 0x5D, 0x5E, 0x5F):  # P,X,],^,_
            b = next_byte()
            while b not in (0x07, 0x1B):
                b = next_byte()
            if b == 0x07:
                return bytes(buffer)
            b = next_byte()
            if b == 0x5C:  # \\
                return bytes(buffer)
            bad_byte(b)

        # Escape Sequence
        # ---------------

        while 0x20 <= b <= 0x2F:
            b = next_byte()
        if 0x30 <= b <= 0x7E:
            return bytes(buffer)
        bad_byte(b)

    # ----------------------------------------------------------------------------------

    def make_raw_request(self, *query: None | int | str) -> None | bytes:
        """
        Make a request to this terminal. This method writes an ANSI escape
        sequence to this terminal as a query and then reads an ANSI escape
        sequence as the response.

        The terminal must have TTYs for input and output. It also must be in
        cbreak mode.
        """
        try:
            return (
                self
                .check_tty()
                .check_cbreak_mode()
                .write_control(*query)
                .flush()
                .read_control()
            )
        except TimeoutError:
            return None

    def parse_textual_response(
        self,
        response: None | bytes,
        prefix: str,
        suffix: str,
    ) -> None | str:
        """
        Parse the terminal's textual response to an ANSI escape query.

        This method converts the response to a string, checks that is starts
        with the prefix and ends with the suffix, and then returns the text
        between prefix and suffix. If the suffix is ST, this method also allows
        BEL, as both are used interchangeably for terminating DSC/OSC.

        If the response is ``None`` or malformed, this method returns ``None``.
        """
        if response is None:
            return None

        r = response.decode('utf8')
        if not r.startswith(prefix):
            return None
        if not r.endswith(suffix):
            if suffix != Ansi.ST or not r.endswith(Ansi.BEL):
                return None
            suffix = Ansi.BEL

        end = -len(suffix) if suffix else len(r)
        return r[len(prefix): end]

    def parse_numeric_response(
        self,
        response: None | bytes,
        prefix: bytes,
        suffix: bytes,
    ) -> list[int]:
        """
        Parse the terminal's numeric response to an ANSI escape query.

        This method checks that the given response starts with the prefix and
        ends with the suffix, splits the bytes between prefix and suffix by
        semicolons, and parses the resulting byte fragments as integers. Empty
        byte fragments are parsed as -1. If the suffix ends with ST, this method
        also allows BEL, as both are used interchangeably for terminating
        DSC/OSC.

        If the response is ``None`` or malformed, this method returns an empty
        list.
        """
        if response is None or not response.startswith(prefix):
            return []

        if not response.endswith(suffix):
            # BEL may stand in for ST
            if not suffix.endswith(RawAnsi.ST):
                return []

            suffix = suffix[:-2] + RawAnsi.BEL
            if not response.endswith(suffix):
                return []

        # Allow for empty suffixes
        end = -len(suffix) if suffix else len(response)
        payload = response[len(prefix): end]

        # Allow for ``:`` separator in addition to ``;``
        params: list[int] = []
        for ps in payload.split(b';'):
            for p in ps.split(b':'):
                params.append(int(p) if p else -1)
        return params

    # ----------------------------------------------------------------------------------

    def _do_request_terminal_version(self) -> None | tuple[str, str]:
        response = self.parse_textual_response(
            self.make_raw_request(Ansi.CSI, '>q'),
            prefix=f'{Ansi.DCS}>|',
            suffix=Ansi.ST,
        )
        if response is None:
            return None

        if response.endswith(')'):
            name, _, version = response[:-1].partition('(')
        else:
            name, _, version = response.partition(' ')

        return name, version

    def request_terminal_identity(self) -> None | tuple[str, str]:
        """
        Request the terminal name and version.

        Since support for the CSI >q escape sequence for querying a terminal for
        its name and version is far from universal, this method employs the
        following strategies:

        1. Use CSI >q escape sequence to query terminal.
        2. Inspect the ``TERMINAL_PROGRAM`` and ``TERMINAL_PROGRAM_VERSION``
           environment variables.
        3. On macOS only, get the bundle identifier from the
           ``__CFBundleIdentifier`` environment variable and then use the
           ``mdfind`` and ``mdls`` command line tools to extract the bundle's
           version.

        If any of these methods is successful, this method normalizes the
        terminal name based on a list of known aliases. That includes bundle
        identifiers for Linux and macOS. It also caches the result and returns
        it for future invocations.

        The terminal must have TTYs for input and output. It also must be in
        cbreak mode.
        """
        if self._identity is False:
            identity = self._do_request_terminal_version()
            if identity:
                self._identity = normalize_terminal_name(identity[0]), identity[1]
            else:
                self._identity = identify_terminal()

        return self._identity

    def request_cursor_position(self) -> None | tuple[int, int]:
        """
        Request the cursor position in (x, y) order from this terminal.

        The terminal must have TTYs for input and output. It also must be in
        cbreak mode.
        """
        parameters = self.parse_numeric_response(
            self.make_raw_request(Ansi.CSI, '6n'),
            prefix=RawAnsi.CSI, suffix=b'R'
        )
        return None if len(parameters) != 2 else (parameters[0], parameters[1])

    def request_batch_mode(self) -> BatchMode:
        """
        Determine the terminal's current batch mode.

        The terminal must have TTYs for input and output. It also must be in
        cbreak mode.
        """
        parameters = self.parse_numeric_response(
            self.make_raw_request(Ansi.CSI, "?2026$p"),
            prefix=b"\x1b[?2026;",
            suffix=b"$y",
        )
        return (
            BatchMode(parameters[0])
            if len(parameters) == 1
            else BatchMode.NOT_SUPPORTED
        )

    def request_active_style(self) -> list[int]:
        """
        Request the terminal's current style settings.

        The returned list contains the corresponding SGR parameters. Terminals
        differ significantly in their support for this query. Since just this
        query would help determine color support levels, that is rather ironic.
        For instance, macOS Terminal.app does not handle the query, whereas
        Visual Studio Code's builtin terminal and iTerm 2 both respond with
        well-formed styles, which are completely wrong in case of Visual Studio
        Code.

        The terminal must have TTYs for input and output. It also must be in
        cbreak mode.
        """
        return self.parse_numeric_response(
            self.make_raw_request(Ansi.DCS, '$qm', Ansi.ST),
            prefix=RawAnsi.fuse(RawAnsi.DCS, b'1$r'),
            suffix=RawAnsi.fuse(b'm', RawAnsi.ST),
        )

    def request_color_support(self) -> None | Fidelity:
        """
        Request the terminal's color support.

        This method uses style queries to test for 24-bit and 8-bit color.

        The terminal must have TTYs for input and output. It also must be in
        cbreak mode. This method resets the current style.
        """
        response = (
            self
            .reset_style()
            .write_control(Ansi.CSI, '31m')
            .write_control(Ansi.CSI, '38;2;6;65;234m')
            .flush()
            .request_active_style()
        )

        self.reset_style().flush()

        if not response:
            return None
        if response == [0, 38, 2, 1, 6, 65, 234]:
            return Fidelity.TwentyFourBit

        response = (
            self
            .write_control(Ansi.CSI, '31m')
            .write_control(Ansi.CSI, '38;5;66m')
            .flush()
            .request_active_style()
        )

        self.reset_style().flush()
        if not response:
            return None
        elif response == [0, 38, 5, 66]:
            return Fidelity.EightBit
        else:
            return Fidelity.Ansi

    # ----------------------------------------------------------------------------------

    def _parse_color(
        self,
        index: int,
        name: str,
        response: None | bytes,
    ) -> Color:
        if response is None:
            raise ValueError(f"no response to request for {name}'s color")

        if index <= 1:
            prefix = f'{Ansi.OSC}{10 + index};'
        else:
            prefix = f'{Ansi.OSC}4;{index - 2};'

        r = self.parse_textual_response(response, prefix=prefix, suffix=Ansi.ST)
        if r is None or not r.startswith('rgb:'):
            raise ValueError(f"malformed response for {name}'s color")

        return Color.parse(r)

    def request_ansi_color(self, color: int) -> Color:
        """
        Determine the color for the given extended ANSI color. This method
        queries the terminal, parses the result, which by convention uses four
        hexadecimal digits per component, and normalizes it to sRGB.

        The terminal must have TTYs for input and output. It also must be in
        cbreak mode.
        """
        assert 0 <= color <= 15

        return self._parse_color(
            color + 2,
            theme.ThemeEntry.try_from_index(color + 2).name(),
            self.make_raw_request(Ansi.OSC, 4, color, ';?', Ansi.ST)
        )

    def request_dynamic_color(self, code: int) -> Color:
        """
        Determine the color for the user interface element identified by
        ``code``:

            * 10 is the foreground or text color
            * 11 is the background color

        This method queries the terminal, parses the result, which by convention
        uses four hexadecimal digits per component, and normalizes it to sRGB.

        The terminal must have TTYs for input and output. It also must be in
        cbreak mode.
        """
        assert 10 <= code <= 11

        return self._parse_color(
            code - 10,
            theme.ThemeEntry.try_from_index(code - 10).name(),
            self.make_raw_request(Ansi.OSC, code, ';?', Ansi.ST),
        )

    def _request_theme_v1(self) -> theme.Theme:
        # (1) Completely process each color.
        colors: list[Color] = []

        for code in range(10, 12):
            colors.append(self.request_dynamic_color(code))
        for code in range(16):
            colors.append(self.request_ansi_color(code))

        return theme.Theme(colors)

    def _request_theme_v2(self) -> theme.Theme:
        # (1) Write all requests. (2) Read + parse all responses.
        colors: list[Color] = []

        self.check_tty().check_cbreak_mode()

        for code in range(10, 12):
            self.write_control(Ansi.OSC, code, ';?', Ansi.ST)
        for color in range(16):
            self.write_control(Ansi.OSC, 4, color, ';?', Ansi.ST)
        self.flush()

        for index in range(18):
            response = self.read_control()
            colors.append(self._parse_color(
                index,
                theme.ThemeEntry.try_from_index(index).name(),
                response,
            ))

        return theme.Theme(colors)

    def _request_theme_v3(self) -> theme.Theme:
        # (1) Write all requests. (2) Read all responses. (3) Parse all responses.
        colors: list[Color] = []

        self.check_tty().check_cbreak_mode()

        for code in range(10, 12):
            self.write_control(Ansi.OSC, code, ';?', Ansi.ST)
        for color in range(16):
            self.write_control(Ansi.OSC, 4, color, ';?', Ansi.ST)
        self.flush()

        responses: list[None | bytes] = []
        for _ in range(18):
            responses.append(self.read_control())

        for index, response in enumerate(responses):
            colors.append(self._parse_color(
                index,
                theme.ThemeEntry.try_from_index(index).name(),
                response,
            ))

        return theme.Theme(colors)

    request_theme = _request_theme_v3
    """
    Request all theme colors from the terminal.

    Currently, there are three different implementations of this method:

     1. The first version completely processes one color at a time. It writes
        the query, then reads the response, and then parses the response.
     2. The second version operates in two phases: It first writes all 18
        queries and then reads and parses all 18 responses.
     3. The third version operates in three phases: It first writes all 18
        queries, then reads all 18 responses, and finally parses all 18
        responses.

    In my measurements, the second and third version take only half the time of
    the first version and, usually, the third version is a bit faster still. But
    I have seen a couple of spurious failures for terminal queries and hence
    expect the need for some retry logic, which would complicate things. So for
    now, I am not ready to commit to either of those three versions. If you feel
    like experimenting, you can run the microbenchmarks by running this module:

    .. code-block:: console

        python -m prettypretty.terminal

    You can also switch between versions by updating the assignment above this
    documentation comment in the source code. In either case, please report back
    about your experiences by `filing an issue
    <https://github.com/apparebit/prettypretty/issues/new/choose>`_.

    The terminal must have TTYs for input and output. It also must be in cbreak
    mode.
    """

    # ----------------------------------------------------------------------------------
    # Terminal Context

    def terminal_theme(self, theme: None | theme.Theme = None) -> TerminalContextManager:
        """
        Use a different color theme. Unless a theme argument is provided, the
        implementation queries the terminal for its current theme, while
        temporarily putting the terminal in cbreak mode.
        """
        return TerminalContextManager(self).terminal_theme(theme)

    def window_title(self, title: str) -> TerminalContextManager:
        """Use a different window title."""
        return TerminalContextManager(self).window_title(title)

    def alternate_screen(self) -> TerminalContextManager:
        """Switch to the terminal's alternate (unbuffered) screen."""
        return TerminalContextManager(self).alternate_screen()

    def hidden_cursor(self) -> TerminalContextManager:
        """Make cursor invisible."""
        return TerminalContextManager(self).hidden_cursor()

    def batched_output(self) -> TerminalContextManager:
        """Batch terminal output."""
        return TerminalContextManager(self).batched_output()

    def bracketed_paste(self) -> TerminalContextManager:
        """
        Enable `bracketed pasting
        <https://gitlab.com/gnachman/iterm2/-/wikis/Paste-Bracketing>`_.
        """
        return TerminalContextManager(self).bracketed_paste()

    def scoped_style(self) -> TerminalContextManager:
        """
        Scope style changes by resetting the style on exit. The ``style()``
        context helps protect against unwanted style leakage upon unexpected
        exceptions or signals.
        """
        return TerminalContextManager(self).scoped_style()

    # ----------------------------------------------------------------------------------

    def up(self, rows: None | int = None) -> Self:
        """Move cursor up."""
        return self.write_control(Ansi.CSI, rows, 'A')

    def down(self, rows: None | int = None) -> Self:
        """Move cursor down."""
        return self.write_control(Ansi.CSI, rows, 'B')

    def left(self, columns: None | int = None) -> Self:
        """Move cursor left."""
        return self.write_control(Ansi.CSI, columns, 'C')

    def right(self, columns: None | int = None) -> Self:
        """Move cursor right."""
        return self.write_control(Ansi.CSI, columns, 'D')

    def at(self, row: None | int = None, column: None | int = None) -> Self:
        """Move the cursor to the given row and column."""
        return self.write_control(Ansi.CSI, row, column, 'H')

    def column(self, column: None | int = None) -> Self:
        """Move the cursor to the given column"""
        return self.write_control(Ansi.CSI, column, 'G')

    def erase_screen(self) -> Self:
        """Erase the entire screen."""
        return self.write_control(Ansi.CSI, '2J')

    def erase_line(self) -> Self:
        """Erase the entire current line."""
        return self.write_control(Ansi.CSI, '2K')

    def link(self, text: str, href: str, id: None | str = None) -> Self:
        """
        `Mark a hyperlink
        <https://gist.github.com/egmontkob/eb114294efbcd5adb1944c9f3cb5feda>`_.

        Underlined text should only be used for hyperlinks, in terminal
        emulators just as much as in documents and on web pages. That's just why
        this class does *not* have a separate method for styling text as
        underlined. If that's too stringent for your use case, please do `open
        an issue <https://github.com/apparebit/prettypretty/issues/new>`_.
        """
        if self._fidelity is Fidelity.Plain:
            self.write(text)
            return self

        code = f"8;id={id};" if id else "8;;"
        return (
            self
            .write_control(Ansi.OSC, code, href, Ansi.ST)
            .write(text)
            .write_control(Ansi.OSC, "8;;", Ansi.ST)
        )

    # ----------------------------------------------------------------------------------

    def is_dark_theme(self) -> bool:
        """Determine whether the current theme is a dark theme."""
        return current_translator().is_dark_theme()

    def reset_style(self) -> Self:
        """Reset all styles."""
        return self.write_control(Ansi.CSI, 'm')

    @overload
    def render(self, __fragments: Sequence[Style|str]) -> Self:
        ...
    @overload
    def render(self, *__fragments: Style | str) -> Self:
        ...
    def render(self, *fragments: Sequence[Style|str] | Style | str) -> Self:
        """Write rich text to terminal output"""
        translator = current_translator()

        if len(fragments) == 1 and isinstance(fragments[0], Sequence):
            ff = fragments[0]
        else:
            ff = fragments

        for fragment in ff:
            if isinstance(fragment, str):
                self.write(fragment)
            else:
                assert isinstance(fragment, Style)
                self.write_control(str(fragment.cap(self._fidelity, translator)))

        return self

    def bold(self) -> Self:
        """Set bold style."""
        return self.write_control(Ansi.CSI, '1m')

    def italic(self) -> Self:
        """Set italic style."""
        return self.write_control(Ansi.CSI, '2m')

    @overload
    def fg(self, color: int, /) -> Self:
        ...
    @overload
    def fg(self, c1: int, c2: int, c3: int, /) -> Self:
        ...
    @overload
    def fg(self, color: Color, /) -> Self:
        ...
    @overload
    def fg(
        self,
        color: int | AnsiColor | EmbeddedRgb | GrayGradient | Rgb | Color | Colorant,
        /
    ) -> Self:
        ...
    def fg(
        self,
        c1: int | AnsiColor | EmbeddedRgb | GrayGradient | Rgb | Color | Colorant,
        c2: None | int = None,
        c3: None | int = None,
    ) -> Self:
        """Set the foreground color."""
        if isinstance(c1, int):
            if c2 is None:
                assert c3 is None
                c1 = Colorant.of(c1)
            else:
                assert c3 is not None
                c1 = Colorant.of(Rgb(c1, c2, c3))
        elif isinstance(c1, Color):
            c1 = Colorant.of(c1)

        translator = current_translator()
        color = translator.cap(c1, self._fidelity)
        if color is not None:
            self.write_control(color.display(Layer.Foreground))
        return self

    @overload
    def bg(self, color: int, /) -> Self:
        ...
    @overload
    def bg(self, c1: int, c2: int, c3: int, /) -> Self:
        ...
    @overload
    def bg(self, color: Color, /) -> Self:
        ...
    @overload
    def bg(
        self,
        color: int | AnsiColor | EmbeddedRgb | GrayGradient | Rgb | Color | Colorant,
        /
    ) -> Self:
        ...
    def bg(
        self,
        c1: int | AnsiColor | EmbeddedRgb | GrayGradient | Rgb | Color | Colorant,
        c2: None | int = None,
        c3: None | int = None,
    ) -> Self:
        """Set the background color."""
        if isinstance(c1, int):
            if c2 is None:
                assert c3 is None
                c1 = Colorant.of(c1)
            else:
                assert c3 is not None
                c1 = Colorant.of(Rgb(c1, c2, c3))
        elif isinstance(c1, Color):
            c1 = Colorant.of(c1)

        translator = current_translator()
        color = translator.cap(c1, self._fidelity)
        if color is not None:
            self.write_control(color.display(Layer.Background))
        return self


if __name__ == '__main__':
    import timeit

    with Terminal().cbreak_mode() as term:
        timer1: Any = timeit.Timer(lambda: term._request_theme_v1())  # type: ignore
        time_taken = timer1.timeit(10)
        print(f'1 stage : {time_taken / 10:.6f}')

        timer2: Any = timeit.Timer(lambda: term._request_theme_v2())  # type: ignore
        time_taken = timer2.timeit(10)
        print(f'2 stages: {time_taken / 10:.6f}')

        timer3: Any = timeit.Timer(lambda: term._request_theme_v3())  # type: ignore
        time_taken = timer3.timeit(10)
        print(f'3 stages: {time_taken / 10:.6f}')



"""Low-level support for assembling ANSI escape sequences"""
import enum


class Ansi(enum.StrEnum):
    """
    An enumeration of ANSI escape sequence components.

    Attributes:
        ESC: is the escape character by itself
        CSI, DCS, OSC: start ANSI escape sequences; they are defined with the
            two-character C0 sequences and not the one-character C1 sequences,
            since the latter conflict with UTF-8
        BEL, ST: are interchangeable and terminate OSC sequences; the latter,
            again, is defined using the two-character C0 sequence.

    All enumeration constants are strings and hence can be directly used when
    assembling ANSI escape sequences. For example:

    .. code-block:: python

        print(f"{Ansi.CSI}1m" "Parrot!" f"{Ansi.CSI}m")

    The example prints a bold excited parrot to the terminal. It leverages
    Python's support for splitting string literals into more than one successive
    literal for clarity and the fact that missing ANSI parameters default to 0
    in the second ANSI escape sequence.

    Since ANSI escape sequences usually are more complex than the above, this
    class also defines :meth:`color_parameters` for converting colors to ANSI
    escape sequence parameters and :meth:`fuse` for fusing all these parts into
    a complete ANSI escape sequence. Using the methods, we can improve on the
    above example as follows:

    .. code-block:: python

        print(
            Ansi.fuse(Ansi.CSI, *Ansi.color_parameters(40), 1, 'm'),
            'Parrot!',
            Ansi.fuse(Ansi.CSI, 'm'),
        )

    The example prints a *green*, bold, excited parrot thanks to the 8-bit color
    40. Fuse not only joins its arguments, but it also inserts semicolons
    between parameters. In other words, the above example prints the same as the
    following statement:

    .. code-block:: python

        print("\\x1b[38;5;40;1m" "Parrot!" "\\x1b[m")

    The example is, again, leveraging Python's support for multiple consecutive
    string literals for improved clarity. Ironically, the Sphinx documentation
    processor mangles an escaped escape character ``\\x1b``, which is now
    written with an escaped backslash in the docstring source. Apparently, there
    is value in hiding escape characters behind enumeration constants...
    """
    BEL = '\a'
    CSI = '\x1b['
    DCS = '\x1bP'
    ESC = '\x1b'
    OSC = '\x1b]'
    ST = '\x1b\\'

    @staticmethod
    def fuse(*fragments: None | int | str) -> str:
        """
        Fuse the ANSI escape sequence fragments into a single string. This
        method treats ``None`` as a default parameter and replaces it with an
        empty string. It also inserts semicolons between parameters, i.e., when
        two successive arguments are either ``None`` or an integer.
        """
        processed: list[str] = []
        previous_was_parameter = False

        for fragment in fragments:
            current_is_parameter = fragment is None or isinstance(fragment, int)
            if previous_was_parameter and current_is_parameter:
                processed.append(';')
            processed.append('' if fragment is None else str(fragment))
            previous_was_parameter = current_is_parameter

        return ''.join(processed)


class RawAnsi(bytes, enum.Enum):
    """
    An enumeration of raw ANSi escape sequence components.

    This is a simpler, ``bytes``-valued version of :class:`Ansi`; see its
    documentation for more details.

    Attributes:
        BEL:
        CSI:
        DCS:
        ST:
    """
    BEL = b'\a'
    CSI = b'\x1b['
    DCS = b'\x1bP'
    ST = b'\x1b\\'

    @staticmethod
    def fuse(*fragments: bytes) -> bytes:
        """Fuse the bytes fragments together."""
        return b''.join(fragments)



"""
Making sense of ANSI colors.
"""
import sys

try:
    import matplotlib.pyplot as plt
    from matplotlib.ticker import FixedFormatter, FixedLocator, FuncFormatter
except ImportError:
    print("prettypretty.plot requires matplotlib. Please install the package,")
    print("e.g., by executing `pip install matplotlib`, and then")
    print("run `python -m prettypretty.plot` again.")
    sys.exit(1)

import argparse
from collections.abc import Sequence
import math
import pathlib
from typing import Any, cast

from .color import (
    close_enough,
    Color,
    ColorSpace,
    spectrum, # pyright: ignore [reportMissingModuleSource]
    theme, # pyright: ignore [reportMissingModuleSource]
)
from .color.gamut import GamutTraversalStep # pyright: ignore [reportMissingModuleSource]
from .progress import add_fidelity, ProgressBar
from .terminal import Terminal
from .theme import current_translator


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="""
            Plot colors on the chroma/hue plane of Oklab while ignoring their
            lightness. If the -i/--input option is specified, this script reads
            newline-separated colors from the named file. If the --vga option is
            specified, it uses the VGA colors. Otherwise, it uses the terminal's
            current ANSI theme colors. This script correctly plots colors
            irrespective of their color space and the current display gamut.
            However, since its visualization library (matplotlib) is limited to
            sRGB, each mark's color is gamut-mapped to sRGB.
        """,
        epilog="""
            Plot accepts a subset of CSS color syntax, including hash
            hexadecimal notation (#123abc) as well as functional notation. For
            the latter, it recognizes "color(<space> <c1> <c2> <c3>)",
            "oklab(<l> <a> <b>)", and "oklch(<l> <c> <h>)". Valid color space
            names are sRGB, linear-sRGB, Display-P3, --linear-Display-P3,
            Rec2020, --linear-Rec2020, --Oklrab, --Oklrch, XYZ, and XYZ-D65 (all
            case-insensitive).
        """
    )
    parser.add_argument(
        "-q", "--quiet",
        action="count",
        default=0,
        help="run silently, without printing status updates"
    )
    parser.add_argument(
        "-v", "--verbose",
        action="count",
        default=0,
        help="run in verbose mode, which prints extra information to the console"
    )
    parser.add_argument(
        "--vga",
        action="store_const",
        const=theme.VGA_COLORS,
        dest="theme",
        help="use VGA colors instead of querying terminal"
    )
    parser.add_argument(
        "-c", "--color",
        action="append",
        dest="colors",
        help="also plot color specified in CSS syntax"
    )

    inputs = parser.add_mutually_exclusive_group()
    inputs.add_argument(
        "-i", "--input",
        help="read newline-separated colors in CSS syntax from named file",
    )
    inputs.add_argument(
        "--oktriples",
        help=(
            "read newline-separated Oklab colors "
            "as comma-separated floating point numbers from named file"
        ),
    )

    parser.add_argument(
        "--no-light",
        action="store_true",
        help="don't include bar chart for lightness",
    )
    parser.add_argument(
        "--no-term",
        action="store_true",
        help=(
            "don't render terminal colors, "
            "only those specified with --color, --input, --oktriples"
        ),
    )
    parser.add_argument(
        "-g", "--gamut",
        action="append",
        dest="gamuts",
        help="also plot boundary of colorspace gamut; value must be sRGB, P3, or Rec2020",
    )
    parser.add_argument(
        "--strong-gamut",
        action="store_true",
        help="use darker, more intense colors for gamut boundaries",
    )
    parser.add_argument(
        "--spectrum",
        action="store_true",
        help="also plot boundary of visible spectrum (experimental)",
    )
    parser.add_argument(
        "--chromaticity",
        action="store_true",
        help="create xy chromaticity diagram"
    )
    parser.add_argument(
        "-o", "--output",
        help="write color plot to the named file"
    )
    return parser


def ingest_oktriples(file: str) -> list[Color]:
    """
    Ingest the file with floating point triplets. Unless the line is empty or
    starts with a hash, it must contain three comma-separated floating point
    values.
    """
    colors: list[Color] = []

    with open(file, mode="r", encoding="utf8") as fd:
        for line in fd:
            line = line.strip()
            if line == "" or line.startswith("#"):
                continue

            color = Color(
                ColorSpace.Oklab,
                [float(num.strip()) for num in line.split(",")]
            )
            colors.append(color)

    return colors


class ColorPlotter:
    ACHROMATIC_THRESHOLD = 0.05

    def __init__(
        self,
        collection_name: None | str = None,
        color_kind: None | str = None,
        segment_size: None | int = None,
        strong_gamut: bool = False,
        chromaticity: bool = False,
        volume: int = 1,
        progress: None | ProgressBar = None,
    ) -> None:
        self._collection_name = collection_name
        self._color_kind = color_kind

        # Scattered colors
        self._xs: list[float] = []
        self._ys: list[float] = []
        self._marks: list[str] = []
        self._mark_colors: list[str] = []

        # Averaged grays as one
        self._grays: list[float] = []
        self._gray_mark = "o"

        # Lightness bars
        self._lightness: list[float] = []
        self._bar_color: list[str] = []
        self._bar_label: list[None | str] = []

        # Counts
        self._colors: set[Color] = set()
        self._duplicate_count = 0
        self._base_count = 0
        self._extra_count = 0

        # Gamut boundaries
        self._segment_size = segment_size or 20
        self._largest_gamut: None | ColorSpace = None
        self._strong_gamut = strong_gamut
        self._with_spectrum = False
        self._chromaticity = chromaticity

        # Spectrum trace
        self._illuminated_observer: None | spectrum.IlluminatedObserver = None
        self._white_point: None | Color = None

        self._volume = volume
        self._progress = progress if 1 <= volume else None
        print(self._progress)

    def status(self, msg: str) -> None:
        if self._volume >= 1:
            print(msg)

    def detail(self, msg: str) -> None:
        if self._volume >= 2:
            print(msg)

    # ----------------------------------------------------------------------------------
    # Individual Color Markers

    def start_adding(self) -> None:
        self.status("                Color    L        Chroma   Hue    Dup")
        self.status("-----------------------------------------------------")

    def add(
        self, name: str, color: Color, marker: str = "o", label: None | str = None
    ) -> None:
        # Matplotlib is sRGB only
        hex_color = color.to_hex_format()

        # Convert to Oklrch
        oklrch = color.to(ColorSpace.Oklrch)
        lr, c, h = oklrch
        c = round(c, 14)  # Chop off one digit of precision.

        # Detect duplicates
        dup = " ✘ " if oklrch in self._colors else " - "

        # Update status
        light = f'{lr:.5}'
        if len(light) > 7:
            light = f'{lr:.5f}'
        chroma = f'{c:.5}'
        if len(chroma) > 7:
            chroma = f'{c:.5f}'
        hue = f'{h:.1f}'
        self.status(f"{name:14}  {hex_color}  {light:<7}  {chroma:<7}  {hue:>5}  {dup}")

        # Skip duplicates
        if oklrch in self._colors:
            self._duplicate_count += 1
            return

        # Display lightness for *all* non-duplicate colors
        self._lightness.append(lr)
        self._bar_color.append(hex_color)
        self._bar_label.append(label)

        # Handle grays
        if oklrch.is_achromatic_threshold(ColorPlotter.ACHROMATIC_THRESHOLD):
            self._grays.append(lr)
            return

        # Add hue/chroma/mark and update counts accordingly
        x, y = self.to_2d(color)
        self._xs.append(x)
        self._ys.append(y)
        self._marks.append(marker)
        self._mark_colors.append(hex_color)
        self._colors.add(oklrch)

        if marker == "o":
            self._base_count += 1
        else:
            self._extra_count += 1

    def stop_adding(self) -> None:
        self.status(
            f"\nTotal: {self._base_count}+{self._extra_count} chromatic colors, "
            f"{len(self._grays)} achromatic colors, "
            f"and {self._duplicate_count} duplicates"
        )

    @property
    def total_count(self) -> int:
        return self._base_count + self._extra_count + len(self._grays)

    def format_counts(self) -> str:
        counts = f"{self._base_count + self._extra_count}"
        grays = len(self._grays)
        if grays > 0:
            counts += f"+{grays}"
        return counts

    # ----------------------------------------------------------------------------------
    # Gamut Tracing

    def trace_spectrum(
        self,
        axes: Any,
        locus_only: bool = False
    ) -> None:
        if self._illuminated_observer is None:
            self._illuminated_observer = spectrum.IlluminatedObserver(
                spectrum.CIE_ILLUMINANT_D65,
                spectrum.CIE_OBSERVER_2DEG_1931,
            )
            self._white_point = self._illuminated_observer.white_point()

        points: tuple[list[float], list[float], list[float], list[str]] = [], [], [], []
        lines2d: list[list[tuple[float, float]]] = []
        all_colors: list[list[str]] = []

        iterator = self._illuminated_observer.visual_gamut(3)
        total_steps = len(iterator)

        self.status(f"Sampling human visual gamut at {total_steps:,} points")
        if self._progress is not None:
            self.status("")

        for index, step in enumerate(iterator):
            if self._progress is not None:
                self._progress.render(index / total_steps * 100)

            if isinstance(step, GamutTraversalStep.MoveTo):
                lines2d.append([])
                all_colors.append([])
                if locus_only and len(lines2d) > 1:
                    break
            c = step.color()
            points[0].append(c[0])
            points[1].append(c[1])
            points[2].append(c[2])
            points[3].append(c.to_hex_format())
            lines2d[-1].append(self.to_2d(c))
            all_colors[-1].append(c.to_hex_format())

        if self._progress is not None:
            self._progress.done()

        if not locus_only:
            f, a = plt.subplots(subplot_kw={"projection": "3d"})  # type: ignore
            a.scatter(points[0], points[1], points[2], c=points[3])  # type: ignore
            f.show()
            #f.savefig("spectrum-3d.svg") # type: ignore
            #plt.close(f) # type: ignore

        total_lines = len(lines2d)

        self.status(f"Connecting gamut samples into {total_lines} lines")
        if self._progress is not None:
            self.status("")

        for index, (line, colors) in enumerate(zip(lines2d, all_colors)):
            if self._progress is not None:
                self._progress.render(index / total_lines * 100)

            self.add_line(
                axes,
                line,
                color=colors, #'#000' if locus_only else '#ccc',
                width=2 if locus_only else 0.8,
            )
        if self._progress is not None:
            self._progress.done()

    def trace_gamut(self, space: ColorSpace, axes: Any) -> None:
        """Trace the boundary of the gamut for the given color space."""
        self.status(f"Tracing {space} gamut")

        if (
            self._largest_gamut is None
            or space is ColorSpace.Rec2020
            or space is ColorSpace.DisplayP3 and self._largest_gamut is ColorSpace.Srgb
        ):
            self._largest_gamut = space

        all_points: list[tuple[float, float]]= []
        all_colors: list[str] = []

        iter = space.gamut(self._segment_size)
        assert iter is not None
        for step in iter:
            color = step.color()

            _, c, h = color.to(ColorSpace.Oklrch)
            if self._strong_gamut:
                hex_format = Color(ColorSpace.Oklrch, [0.6, c, h]).to_hex_format()
            else:
                hex_format = Color(ColorSpace.Oklrch, [0.75, c / 3, h]).to_hex_format()

            pt = self.to_2d(color)

            self.detail(
                f"{color.space()} gamut: {color[0]:.2f}, {color[1]:.2f}, {color[2]:.2f} --> "
                f"{self.format_2d(*pt)} ({hex_format})"
            )

            all_points.append(pt)
            all_colors.append(hex_format)

            if isinstance(step, GamutTraversalStep.CloseWith):
                break

        assert close_enough(all_points[0][0], all_points[-1][0])
        assert close_enough(all_points[0][1], all_points[-1][1])

        self.add_line(axes, all_points, all_colors)

    def add_gamut_label(
        self,
        axes: Any,
        space: ColorSpace,
        label: str,
        green: float,
        blue: float,
        dtext: float,
        pad: int,
    ) -> None:
        x, y = self.to_2d(Color(space, [0.0, green, blue]))
        axes.annotate(  # type: ignore
            label, xy=(x, y), xytext=(x, y + dtext), textcoords="data",
            bbox=dict(
                pad=pad,
                facecolor="none",
                edgecolor="none"
            ),
            arrowprops=dict(
                arrowstyle="-",
                connectionstyle="arc3",
                shrinkA=0,
                shrinkB=0,
            ),
        )

    def add_line(
        self,
        axes: Any,
        points: Sequence[tuple[float, float]],
        color: str | Sequence[str],
        width: float = 1.5,
    ) -> None:
        should_index_color = not isinstance(color, str)

        x0: None | float = None
        y0: None | float = None
        for index, (x1, y1) in enumerate(points):
            if x0 is not None:
                assert y0 is not None
                c = color[index] if should_index_color else color

                axes.plot(
                    [x0, x1],
                    [y0, y1],
                    c = c,
                    lw = width,
                )

            x0 = x1
            y0 = y1

    def to_2d(self, color: Color) -> tuple[float, float]:
        return color.xy_chromaticity() if self._chromaticity else color.hue_chroma()

    def format_2d(self, x: float, y: float) -> str:
        if not self._chromaticity:
            return f'h={x:.5f}  c={y:.5f}'
        else:
            return f'x={x:.5f}  y={y:.5f}'

    # ----------------------------------------------------------------------------------
    # Figure Creation

    def create_figure(
        self,
        collection_name: None | str = None,
        color_kind: None | str = None,
        gamuts: None | list[ColorSpace] = None,
        with_spectrum: bool = False,
        with_lightness: bool = True,
    ) -> Any:
        self.status("Creating figure")

        if self._chromaticity:
            fig: Any = plt.figure(figsize=(5, 5.5)) # type: ignore
            axes: Any = fig.add_subplot()
            light_axes: Any = None
        elif with_lightness:
            fig = plt.figure(layout="constrained", figsize=(5, 6.5))  # type: ignore
            axes = fig.add_subplot(6, 10, (1, 50), polar=True)
            light_axes = fig.add_subplot(6, 10, (51, 60))
        else:
            fig = plt.figure(layout="constrained", figsize=(5, 5.5)) # type: ignore
            axes = fig.add_subplot(polar=True)
            light_axes = None

        # Add spectrum and gamut boundaries if so requested.
        # if with_spectrum or self._chromaticity:
        #     self.premultiply(spectrum.CIE_OBSERVER_2DEG_1931, spectrum.CIE_ILLUMINANT_D65)
        if with_spectrum:
            self._with_spectrum = with_spectrum
            self.trace_spectrum(axes)
        if self._chromaticity:
            self.trace_spectrum(axes, locus_only=True)

            # Add white point
            assert self._white_point is not None
            x, y = self.to_2d(self._white_point)
            self._xs.append(x)
            self._ys.append(y)
            self._marks.append("*")
            self._mark_colors.append(self._white_point.to_hex_format())

        gamuts = gamuts or []
        for space in gamuts:
            self.trace_gamut(space, axes)

        # Since markers are shared for all marks in a series, we use a new
        # series for every single color.
        total_markers = len(self._xs)

        self.status(f"Setting {total_markers:,} chromatic markers")
        if self._progress is not None and 100 <= total_markers:
            self.status("")

        for index, (x, y, color, marker) in enumerate(zip(
            self._xs, self._ys, self._mark_colors, self._marks
        )):
            if self._progress is not None and 100 <= total_markers:
                self._progress.render(index / total_markers * 100)

            if marker == "o":
                size = 80
            elif marker == "*":
                size = 120
            else:
                size = 60
            axes.scatter(
                [x],
                [y],
                c=[color],
                s=[size],
                marker=marker,  # type: ignore
                edgecolors='#000',
                zorder=5,
            )
        if self._progress is not None and 100 <= total_markers:
            self._progress.done()

        if self._grays and not self._chromaticity:
            gray = Color.oklab(sum(self._grays) / len(self._grays), 0.0, 0.0).to_hex_format()

            axes.scatter(
                [0],
                [0],
                c=[gray],
                s=[80],
                marker=self._gray_mark,
                edgecolors='#000',
            )

        if self._chromaticity:
            axes.set_xlim([0, 0.8])
            axes.set_ylim([0, 0.9])
            axes.grid()

        else:
            axes.set_aspect(1)
            axes.set_rmin(0)
            axes.set_rmax(self.effective_max_chroma())

            # Don't show tick labels at angle
            axes.set_rlabel_position(0)

            # When max chroma is 0.3 or 0.4, matplotlib generates grid circles every
            # 0.05 units and, for larger max chroma, every 0.1 units. Always
            # generate labels every 0.1 units, but shift them by 0.05 in the former
            # case. Never generate labels for origin or max chroma.
            axes.yaxis.set_major_formatter(FuncFormatter(
                    self.format_ytick_label_10 if self.effective_max_chroma() >= 0.5
                    else self.format_ytick_label_05
            ))

            # Center tick labels on tick
            plt.setp(axes.yaxis.get_majorticklabels(), ha="center")  # type: ignore

        # Make grid appear below points
        axes.set_axisbelow(True)

        lry_offset = 0
        if not self._chromaticity:
            # Add label for gamut boundaries
            if ColorSpace.Srgb in gamuts:
                self.add_gamut_label(axes, ColorSpace.Srgb, "sRGB", 1.0, 0.85, 0.12, 0)
            if ColorSpace.DisplayP3 in gamuts:
                self.add_gamut_label(axes, ColorSpace.DisplayP3, "P3", 1.0, 0.95, 0.07, 0)
            if ColorSpace.Rec2020 in gamuts:
                self.add_gamut_label(axes, ColorSpace.Rec2020, "Rec. 2020", 0.95, 1.0, 0.18, 1)

        if not self._chromaticity and with_lightness:
            light_axes.set_yticks([0, 0.5, 1], minor=False)
            light_axes.set_yticks([0.25, 0.75], minor=True)
            light_axes.yaxis.grid(True, which="major")
            light_axes.yaxis.grid(True, which="minor")

            light_axes.bar(
                [x for x in range(len(self._lightness))],
                self._lightness,
                color=self._bar_color,
                zorder=5,
                edgecolor="#000",
                linewidth=0.5,
            )

            light_axes.set_ylim(0, 1)
            light_axes.margins(x=0.02, tight=True)

            if all(label is not None for label in self._bar_label):
                lry_offset = 0.015
                light_axes.xaxis.set_major_locator(
                    FixedLocator([*range(len(self._bar_label))])
                )
                light_axes.xaxis.set_major_formatter(
                    FixedFormatter(self._bar_label) # pyright: ignore [reportArgumentType]
                )
            else:
                light_axes.get_xaxis().set_visible(False)

        collection_name = collection_name or self._collection_name
        color_kind = color_kind or self._color_kind

        title = f"{collection_name}: " if collection_name else ""
        title += f"{self.format_counts()} "
        if color_kind:
            title += f"{color_kind} "
        title += "Color"
        if self.total_count != 1 or self._base_count != 1:
            title += "s"

        title += "" if self._chromaticity else " in Oklab"
        if self._chromaticity:
            fig.suptitle("CIE 1931 xy Chromaticity", weight="bold", size=13)
            axes.set_title(title, style="italic", size=13)
        else:
            fig.suptitle(title, ha="left", x=0.044, weight="bold", size=13)
            axes.set_title("Hue & Chroma", style="italic", size=13, x=0.11, y=1.01)
            if with_lightness:
                fig.text(
                    0.09, 0.185 + lry_offset,
                    "Lightness (Lr)",
                    fontdict=dict(style="italic", size=13)
                )

        return fig

    def effective_max_chroma(self) -> float:
        if self._with_spectrum:
            return 0.5
        elif self._largest_gamut in (ColorSpace.Srgb, ColorSpace.DisplayP3):
            return 0.4
        elif self._largest_gamut is ColorSpace.Rec2020:
            return 0.5
        elif all(c < 0.3 for c in self._ys):
            return 0.3
        else:
            return 0.4

    def format_ytick_label_05(self, y: float, _: int) -> str:
        if y % 0.1 < 1e-9 or math.isclose(y, self.effective_max_chroma()):
            return ""
        else:
            return f"{y:.2}"

    def format_ytick_label_10(self, y: float, _: int) -> str:
        if y > 0.01 and y % 0.1 < 1e-9 and y < self.effective_max_chroma() - 0.01:
            return f"{y:.2}"
        else:
            return ""

# ======================================================================================

def main(options: Any, term: Terminal) -> None:
    gamuts: list[ColorSpace] = []
    for gamut in cast(list[str], options.gamuts or []):
        gamut = gamut.casefold()
        if gamut == 'srgb':
            gamuts.append(ColorSpace.Srgb)
        elif gamut == 'p3':
            gamuts.append(ColorSpace.DisplayP3)
        elif gamut == 'rec2020':
            gamuts.append(ColorSpace.Rec2020)
        else:
            raise ValueError(f"invalid gamut name {gamut}")

    # ----------------------------------------------------------------------------------
    # Prepare Colors for Plotting

    plotter = ColorPlotter(
        volume=1-options.quiet+options.verbose,
        strong_gamut=options.strong_gamut,
        chromaticity=options.chromaticity,
        progress=ProgressBar(term),
    )
    terminal_id = None

    plotter.start_adding()

    if not options.no_term:
        if not options.theme:
            terminal_id = term.request_terminal_identity()

        translator = current_translator()
        for base_index in [0, 1, 3, 2, 6, 4, 5, 7]:
            for index in [base_index, base_index + 8]:
                color = translator.resolve(index)
                name = theme.ThemeEntry.try_from_index(index + 2).name()
                label = theme.ThemeEntry.try_from_index(index + 2).abbr()
                plotter.add(name, color, label=label)

    cname = "" if options.no_term else "<extra>"
    marker = "o" if options.no_term else "d"

    if options.input is not None:
        with open(options.input, mode="r", encoding="utf8") as file:
            for color in [Color.parse(line) for line in file.readlines() if line.strip()]:
                plotter.add(cname, color, marker=marker)
    elif options.oktriples is not None:
        for color in ingest_oktriples(options.oktriples):
            plotter.add(cname, color, marker=".")

    for color in [Color.parse(c) for c in cast(list[str], options.colors) or []]:
        plotter.add(cname, color, marker=marker)

    plotter.stop_adding()

    # ----------------------------------------------------------------------------------
    # Labels and file names

    if options.no_term:
        collection_name = None
    elif options.theme:
        collection_name = "VGA"
    elif terminal_id:
        collection_name = terminal_id[0]
    else:
        collection_name = "Unknown Terminal"

    if options.no_term:
        color_kind = None
    else:
        color_kind = "ANSI"

    if options.output is not None:
        file_name = options.output
    elif options.input is not None:
        file_name = pathlib.Path(options.input).with_suffix(".svg")
    elif options.oktriples is not None:
        file_name = pathlib.Path(options.oktriples).with_suffix(".svg")
    else:
        assert collection_name is not None
        file_name = f'{collection_name.replace(" ", "-").lower()}-colors.svg'

    # ----------------------------------------------------------------------------------
    # Create and save plot

    fig = plotter.create_figure(
        collection_name=collection_name,
        color_kind=color_kind,
        gamuts=gamuts,
        with_spectrum=options.spectrum,
        with_lightness=not options.no_light
    )
    plotter.status(f"Saving plot to `{file_name}`")
    fig.savefig(file_name, bbox_inches="tight")  # type: ignore
    plotter.status("Happy, happy, joy, joy!")


if __name__ == "__main__":
    options = add_fidelity(create_parser()).parse_args()
    with (
        Terminal(options.fidelity)
        .cbreak_mode()
        .terminal_theme(options.theme)
        .hidden_cursor()
        .scoped_style()
    ) as term:
        main(options, term)



"""Utility module for traversing the boundary of a color space's gamut."""

from typing import Self

from . import Color

class GamutTraversalStep_MoveTo(GamutTraversalStep):
    """Start new path by moving to color coordinates."""


class GamutTraversalStep_LineTo(GamutTraversalStep):
    """Continue path by drawing line to color coordinates."""


class GamutTraversalStep_CloseWith(GamutTraversalStep):
    """Close path by drawing line to color coordinates."""


class GamutTraversalStep:
    """A step along a path during gamut boundary traversal."""
    MoveTo = GamutTraversalStep_MoveTo
    LineTo = GamutTraversalStep_LineTo
    CloseWith = GamutTraversalStep_CloseWith

    def color(self) -> Color: ...
    def __repr__(self) -> str: ...


class GamutTraversal:
    """An iterator over RGB gamut boundaries."""
    def __len__(self) -> int: ...
    def __iter__(self) -> Self: ...
    def __next__(
        self
    ) -> GamutTraversalStep.MoveTo | GamutTraversalStep.LineTo | GamutTraversalStep.CloseWith: ...
    def __repr__(self) -> str: ...



"""Utility module for visualizing the human visual gamut."""

from typing import Self

from .. import Color
from ..gamut import GamutTraversalStep
from . import std_observer as std_observer


class Illuminant:
    """
    A standard illuminant. Note that `at` accepts a nanometer wavelength whereas
    `__getitem__` accepts a zero-based index.
    """
    def label(self) -> str: ...
    def is_empty(self) -> bool: ...
    def start(self) -> int: ...
    def end(self) -> int: ...
    def len(self) -> int: ...
    def at(self, wavelength: int) -> float: ...
    def checksum(self) -> float: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> float: ...
    def __repr__(self) -> str: ...


class Observer:
    """
    A standard observer, that is, a color matching function. Note that `at`
    accepts a nanometer wavelength whereas `__getitem__` accepts a zero-based
    index.
    """
    def label(self) -> str: ...
    def is_empty(self) -> bool: ...
    def start(self) -> int: ...
    def end(self) -> int: ...
    def len(self) -> int: ...
    def at(self, wavelength: int) -> list[float]: ...
    def checksum(self) -> list[float]: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> list[float]: ...
    def __repr__(self) -> str: ...


class IlluminatedObserver:
    """
    A table of weighting factors resulting from multiplying the
    per-wavelength values of an illuminant and an observer.
    """
    def __new__(cls, illuminant: Illuminant, observer: Observer) -> Self: ...
    def label(self) -> str: ...
    def is_empty(self) -> bool: ...
    def start(self) -> int: ...
    def end(self) -> int: ...
    def len(self) -> int: ...
    def at(self, wavelength: int) -> list[float]: ...
    def minimum(self) -> list[float]: ...
    def maximum(self) -> list[float]: ...
    def checksum(self) -> list[float]: ...
    def luminosity(self) -> float: ...
    def white_point(self) -> Color: ...
    def visual_gamut(self, stride: int) -> SpectrumTraversal: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> list[float]: ...
    def __repr__(self) -> str: ...


class SpectrumTraversal:
    """An iterator over spectrum boundaries."""
    def stride(self) -> int: ...
    def line_count(self) -> int: ...
    def line_length(self) -> int: ...
    def pulse(self, start: int, width: int) -> list[float]: ...
    def pulse_color(self, start: int, width: int) -> Color: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Self: ...
    def __next__(self) -> GamutTraversalStep.MoveTo | GamutTraversalStep.LineTo: ...
    def __repr__(self) -> str: ...


CIE_ILLUMINANT_D50: Illuminant = ...
CIE_ILLUMINANT_D65: Illuminant = ...
CIE_ILLUMINANT_E: Illuminant = ...
CIE_OBSERVER_2DEG_1931: Observer = ...
CIE_OBSERVER_10DEG_1964: Observer = ...

ONE_NANOMETER: int = ...



import argparse
from collections.abc import Iterator
import random
import time

from prettypretty.color import Color
from prettypretty.color.style import Style # pyright: ignore [reportMissingModuleSource]
from prettypretty.color.termco import (Colorant, Rgb) # pyright: ignore [reportMissingModuleSource]
from prettypretty.terminal import Terminal
from prettypretty.theme import current_translator


def add_fidelity(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Add command line arguments to control use of colors."""
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--nocolor', action='store_const', const='nocolor', dest='fidelity',
        help='do not use any colors'
    )
    group.add_argument(
        '--ansi', action='store_const', const='ansi', dest='fidelity',
        help='use at most ANSI colors only',
    )
    group.add_argument(
        '--eight-bit', action='store_const', const='eight_bit', dest='fidelity',
        help='use at most 8-bit colors',
    )
    return parser


class ProgressBar:
    """A reusable progress bar."""
    BLOCKS = ' ▎▌▊█'  # <empty> <partial>+ <full>
    STEPS = len(BLOCKS) - 1
    WIDTH = 100 // STEPS + (1 if 100 % STEPS != 0 else 0)
    assert WIDTH * STEPS >= 100  # Without the adjustment, this wouldn't hold

    LIGHT_MODE_BAR = Style().with_foreground(Color.p3(0.0, 1.0, 0.0))
    DARK_MODE_BAR = Style().with_foreground(Rgb(3, 151, 49))

    def __init__(self, term: Terminal) -> None:
        """Select light/dark mode style and adjust color."""
        style = self.DARK_MODE_BAR if term.is_dark_theme() else self.LIGHT_MODE_BAR
        self._style = style.cap(term.fidelity, current_translator())
        self._term = term
        self._percent = 0

    @property
    def color(self) -> None | Colorant:
        """Access the effective progress bar color."""
        return self._style.foreground()

    def _format(self, percent: float) -> list[Style| str]:
        """Generate progress bar for given percentage."""
        percent = min(percent, 100)  # Clamp max at 100.0

        # Need integer multiple (full) and index (partial), hence must round
        full, partial = divmod(round(percent), self.STEPS)
        bar = self.BLOCKS[-1] * full
        if partial > 0:
            # Only add partial character if it is non-empty
            bar += self.BLOCKS[partial]
        bar = bar.ljust(self.WIDTH, self.BLOCKS[0])

        # Displayed percentage remains nicely floating point
        return ['  ┫', self._style, bar, -self._style, '┣', f' {percent:5.1f}%']

    def render(self, percent: float) -> None:
        """Render the progress bar for the given percentage."""
        # Ensure empty/full bar renders, throttle rendering in between
        if (self._percent == 0) or self._percent + 1 < percent or (100 <= percent):
            self._percent = percent
            self._term.column(0).render(self._format(percent)).flush()

    def done(self) -> None:
        """Complete progress bar, write newline, and reset internal state."""
        self.render(100)
        self._term.writeln('\n').flush()
        self._percent = 0


def progress_reports() -> Iterator[float]:
    percent = 0.0
    while True:
        yield percent
        if percent >= 100.0:
            return
        percent += random.gauss(1, 1/3)


def main() -> None:
    options = add_fidelity(
        argparse.ArgumentParser("simulate progress bar")
    ).parse_args()

    with (
        Terminal(fidelity=options.fidelity)
        .terminal_theme()
        .hidden_cursor()
        .scoped_style()
    ) as term:
        progress = ProgressBar(term)
        term.writeln(f'Using {progress.color} as color!\n').flush()

        for percent in progress_reports():
            progress.render(percent)
            time.sleep(random.uniform(1/60, 1/10))
        progress.done()


if __name__ == '__main__':
    main()



import os
import sys


def is_dark_mode() -> None | bool:
    """
    Determine whether the operating system is in dark mode.

    Returns:
        ``True`` for dark mode, ``False`` for light mode, and ``None`` if the
        mode could not be determined.

    The implementation builds on answers to `this StackOverflow question
    <https://stackoverflow.com/questions/65294987/detect-os-dark-mode-in-python>`_
    and `the darkdetect package
    <https://github.com/albertosottile/darkdetect>`_. The latter seems both
    over- and under-engineered. In contrast, this module provides the one
    interesting bit, whether the system is in dark mode, if available and
    nothing else.
    """
    try:
        if sys.platform == "darwin":
            return _is_darkmode_macos()
        elif sys.platform == "linux":
            return _is_darkmode_linux()
        elif os.name == 'nt':
            return _is_darkmode_windows()
        else:
            return None
    except:
        return None


def _is_darkmode_windows() -> bool:
    import winreg

    with winreg.OpenKey(  # type: ignore[attr-defined]
        winreg.HKEY_CURRENT_USER,  # type: ignore[attr-defined]
        "Software\\Microsoft\\Windows\\CurrentVersion\\Themes\\Personalize",
    ) as key:  # type: ignore
        return not winreg.QueryValueEx(  # type: ignore[attr-defined]
            key, "AppsUseLightTheme"
        )[0]


def _is_darkmode_macos() -> bool:
    import subprocess

    # Use DEVNULL so that output of command isn't shown
    return not subprocess.run(
        ["defaults", "read", "-g", "AppleInterfaceStyle"],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    ).returncode


def _is_darkmode_linux() -> None | bool:
    import subprocess

    try:
        result = subprocess.run( [
                "dbus-send",
                "--session",
                "--print-reply=literal",
                "--dest=org.freedesktop.portal.Desktop",
                "/org/freedesktop/portal/desktop",
                "org.freedesktop.portal.Settings.Read",
                "string:org.freedesktop.appearance",
                "string:color-scheme"
            ],
            capture_output=True,
            encoding='utf8',
            check=True,
        )

        stdout = result.stdout.replace("variant", "").replace("uint32", "").strip()
        if stdout in ("0", "1", "2"):
            # 0 stands for default, 1 for prefers-dark, and 2 for prefers-light.
            # Ubuntu returns 0 for light mode and 1 for dark mode.
            return stdout == "1"
    except:
        pass

    try:
        result = subprocess.run(
            ["gsettings", "get", "org.gnome.desktop.interface", "color-scheme"],
            capture_output=True,
            encoding="utf8",
            check=True,
        )
        stdout = result.stdout.strip()
    except:
        stdout = ""

    if not stdout:
        result = subprocess.run(
            ["gsettings", "get", "org.gnome.desktop.interface", "gtk-theme"],
            capture_output=True,
            encoding="utf8",
            check=True,
        )
        stdout = result.stdout.strip()
    return stdout.strip("'").endswith("-dark")



import argparse
from collections import Counter
import io
import math
import sys
from textwrap import dedent, indent
from typing import Any, Self

from prettypretty.color import Color, ColorSpace
from prettypretty.color.gamut import ( # pyright: ignore [reportMissingModuleSource]
    GamutTraversalStep
)
from prettypretty.color.spectrum import ( # pyright: ignore [reportMissingModuleSource]
    CIE_ILLUMINANT_D50, CIE_ILLUMINANT_D65, CIE_ILLUMINANT_E, CIE_OBSERVER_2DEG_1931,
    CIE_OBSERVER_10DEG_1964, IlluminatedObserver
)


if sys.stderr.isatty():
    BOLD = "\x1b[1m"
    RESET = "\x1b[m"
else:
    BOLD = "" # pyright: ignore [reportConstantRedefinition]
    RESET = "" # pyright: ignore [reportConstantRedefinition]


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=indent(dedent(f"""
        {BOLD} Visualization of Human Visual Gamut {RESET}

        Generate a point cloud or mesh for the outer volume of the human visual
        gamut in the XYZ color space and store as a PLY file in the
        `visual-gamut` directory.

        Each point's color is the sRGB color resulting from gamut mapping the
        original XYZ color. In other words, colors are far from accurate.

        The best possible resolution, a stride of 1nm, produces a clean shape
        with good coverage of the entire gamut's surface. However, it also
        generates too many vertices (and mesh triangles) for parts of the shape,
        suggesting the need for mesh simplification. Experiments to identify a
        strategy yielding good results are on-going.


        {BOLD}Mesh Generation{RESET}

        This script uses the standard algorithm for generating points: It shifts
        and rotates pulses of increasing width across the visible spectrum. Mesh
        generation is based on two observations:

         1. Points resulting from a pulse with the same width can be arranged on
            a line. Every line has the same number of points.
         2. Lines for subsequent pulse widths traverse largely similar paths in
            3D space shifted by some (varying) distance.

        Mesh generation leverages that structure as follows:

         3. By treating subsequent lines as ropes of a rope ladder, generating
            virtual quads for much of the gamut body becomes straight-forward.
            In practice, the quads are divided into two triangles.
         4. The first and last line leave holes roughly shaped like spoons, with
            a distinct handle and bowl. The point marking the transition is the
            cusp.

              - Close the handle by treating both sides as ropes for a rope
                ladder again. This requires that the cusp's offset from the
                first line point is not divisible by two (since offset are
                zero-based).
              - Mesh the bowl as a fan of triangles. This probably can also be
                conceptualized as another rope ladder, but a fan of triangles
                seems simpler.

         5. Add two capstone triangles, which don't quite fit into the fan
            generation.

        Filling the holes holes remaining after step 4 was easy. But identifying
        the holes as simple triangles was much harder. Towards that end, mesh
        generation keeps track of edge counts and reports any remaining
        boundaries. There are none, but boundary detection remains active.
        """), "    ")
    )
    parser.add_argument(
        "--ok",
        action="store_true",
        help="render gamut in the Oklrab color space instead of XYZ"
    )
    parser.add_argument(
        "-i", "--illuminant",
        default="D65",
        choices=["D50", "d50", "D65", "d65", "E", "e"],
        help="choose between the CIE's D50, D65, and E illuminants",
    )
    parser.add_argument(
        "-b", "--observer",
        default="2",
        choices=["2", "10"],
        help="choose between the CIE's 1931 2º or the 1964 10º standard observer",
    )
    parser.add_argument(
        "-s", "--stride",
        help="use stride s measured in integral nanometers, with 1 <= s <= 20"
    )
    parser.add_argument(
        "-m", "--mesh",
        action="store_true",
        help="generate a closed, manifold, triangular mesh covering the entire visual "
        "gamut",
    )
    parser.add_argument(
        "--darken",
        action="store_true",
        help="darken vertex colors by reducing their lightness in Oklrab to 90%%"
    )
    parser.add_argument(
        "--alpha",
        default="1.0",
        help="make vertex colors transparent, using given alpha between 0 and 1",
    )
    parser.add_argument(
        "-r", "--render",
        action="store_true",
        help="interactively render the gamut in 3D with projected silhouettes; "
        "requires the Vedo 3D library"
    )
    parser.add_argument(
        "-v", "--verbose",
        action="store_true",
        help="run in verbose mode"
    )
    parser.add_argument(
        "--gamut",
        choices=["sRGB", "srgb", "P3", "p3", "Rec2020", "rec2020"],
        help="also plot boundary of gamut for given color space",
    )
    parser.add_argument(
        "--planar-gamut",
        action="store_true",
        help="plot gamut boundary on X/Z plane for XYZ and a/b plane for Oklrab, "
        "with Y=0 or Lr= 0, respectively"
    )

    return parser


def log(msg: str = "") -> None:
    print(msg, file=sys.stderr)


trace_enabled = False
def trace(msg: str = "") -> None:
    if trace_enabled:
        print(msg, file=sys.stderr)


class Sampler:
    def __init__(self) -> None:
        impossible = Color(ColorSpace.Oklrch, [1.0, math.inf, 0.0])
        black = Color(ColorSpace.Oklrch, [0.0, 0.0, 0.0])
        self.minima = [impossible] * 8
        self.maxima = [black] * 8
        self.achromatic = 0
        self.chromatic = 0

    def sample(self, color: Color) -> None:
        color = color.to(ColorSpace.Oklrch)
        if color.is_achromatic():
            self.achromatic += 1
            return

        self.chromatic += 1
        _, chroma, hue = color
        index = int(hue // 45)
        if chroma < self.minima[index][1]:
            self.minima[index] = color
        elif chroma > self.maxima[index][1]:
            self.maxima[index] = color

    def __str__(self) -> str:
        lines: list[str] = []

        lines.append("Hue bin   Lr      Min C   h      Lr      Max C   h")
        lines.append("------------------------------------------------------")

        for index in range(8):
            mini = self.minima[index]
            maxi = self.maxima[index]

            lines.append(
                f"{index * 45:3}-{(index+1) * 45:3}º  "
                f"{mini[0]:.5f} {mini[1]:.5f} {mini[2]:5.1f}  "
                f"{maxi[0]:.5f} {maxi[1]:.5f} {maxi[2]:5.1f}"
            )

        lines.append("")

        chromatic = f"{self.chromatic:9,}"
        chromatic_label = "chromatic".rjust(len(chromatic))
        achromatic = f"{self.achromatic:10,}"
        achromatic_label = "achromatic".rjust(len(achromatic))
        total = f"{self.chromatic + self.achromatic:5,}"
        total_label = "total".rjust(len(total))

        lines.append(f"{total} = {chromatic} + {achromatic} samples")
        lines.append(f"{total_label} = {chromatic_label} + {achromatic_label}")

        return "\n".join(lines)


class PointManager:
    def __init__(
        self,
        *,
        stride: int,
        space: ColorSpace,
        darken: bool = False,
        planar: bool = False,
        mesh: bool = False,
        alpha: float = 1.0,
        label: None | str = None,
    ) -> None:
        self._stride = stride
        self._space = space
        self._should_darken = darken
        self._should_project_to_plane = planar
        self._should_generate_mesh = mesh
        self._points: list[tuple[float, float, float]] = []
        self._colors: list[bytes] = []
        self._faces: list[tuple[int, int, int]] = []
        self._alpha: int = min(max(int(alpha * 255), 0), 255)
        self._line_count = 0
        self._line_length = 0
        self._label = label

    @property
    def line_count(self) -> int:
        return self._line_count

    @property
    def line_length(self) -> int:
        return self._line_length

    @property
    def line_cusp(self) -> int:
        cusp = self.line_length * 32 // 100
        if cusp % 2 == 0:
            cusp += 1
        return cusp

    def __len__(self) -> int:
        return len(self._points)

    @property
    def point_count(self) -> int:
        return len(self._points)

    @property
    def face_count(self) -> int:
        return len(self._faces)

    def mark_newline(self) -> None:
        self._line_count += 1
        if self._line_count == 2:
            self._line_length = self.point_count

    def add(self, *, color: Color, is_xyz: bool, highlight: None | Color = None) -> None:
        x = color[0]
        y = color[1]
        z = color[2]

        if self._should_project_to_plane:
            if is_xyz:
                # Zero out luminosity
                self._points.append((x, 0, z))
            else:
                # Zero out lightness
                self._points.append((0, y, z))
        else:
            self._points.append((x, y, z))

        if highlight is not None:
            self._colors.append(highlight.to_24bit())
            return

        if self._should_darken:
            lr, c, h = color.to(ColorSpace.Oklrch)
            color = Color.oklrch(0.9 * lr, c, h)

        self._colors.append(color.to_24bit())

    def generate_faces(self) -> None:
        # See description in tool help!

        if not self._should_generate_mesh:
            return

        line_count = self.line_count
        line_length = self.line_length
        edges = Counter[tuple[int, int]]()

        # Create rope ladder of quads from series for pulse widths n and n+1.
        # E.g., at 1nm resolution, for n=0, that includes the edge (0, 471).
        for line_index in range(line_count - 1):
            if trace_enabled:
                trace("-" * 141)

            for index in range(line_length - 1):
                s = line_index * line_length + index
                t = s + 1
                u = t + line_length
                v = u - 1

                self._faces.append((s, v, u))
                self._faces.append((s, u, t))

                msg = ""
                if trace_enabled:
                    msg = f"pulse {line_index:3} @ {index:3}"

                for edge in [(s, v), (v, u), (s, u), (s, u), (t, u), (s, t)]:
                    assert edge[0] <= edge[1]
                    edges[edge] += 1

                    if trace_enabled:
                        msg += f" | {edge[0]:7,} -> {edge[1]:7,}"

                trace(msg)

                # Connect end of rope ladder to start of next rope ladder.
                if index == line_length - 2 and line_index < line_count - 2:
                    s = t + 1
                    v = u + 1

                    self._faces.append((t, u, v))
                    self._faces.append((t, v, s))

                    msg = ""
                    if trace_enabled:
                        msg = f"cross {line_index:3} @ {index:3}"

                    for edge in [(s, v), (u, v), (s, u), (s, u), (t, u), (t, s)]:
                        assert edge[0] <= edge[1]
                        edges[edge] += 1

                        if trace_enabled:
                            msg += f" | {edge[0]:7,} -> {edge[1]:7,}"

                    trace(msg)

        # For first & last line, that leaves a hole roughly shaped like a bent
        # spoon. The spoon's handle transitions to the bowl at about 32% of the
        # line, i.e., at the cusp.
        line_cusp = self.line_cusp
        halfcusp = line_cusp // 2  # Rounds downward

        for base in (0, (line_count - 1) * line_length):
            # Fill spoon's handle with another rope ladder of quads.
            if trace_enabled:
                trace("-" * 141)

            for index in range(1, halfcusp + 1):
                # s,t are growing from base,u,v are shrinking from base + line_cusp
                s = base + index - 1
                t = s + 1
                u = base + line_cusp - index
                v = u + 1

                self._faces.append((s, v, u))
                self._faces.append((s, u, t))

                msg = ""
                if trace_enabled:
                    label = "#1" if base == 0 else "#2"
                    msg = f"patch {label}  @ {index:3}"

                for edge in [(s, v), (u, v), (s, u), (s, u), (t, u), (s, t)]:
                    assert edge[0] <= edge[1]
                    edges[edge] += 1

                    if trace_enabled:
                        msg += f" | {edge[0]:7,} -> {edge[1]:7,}"

                trace(msg)

            # Fill spoon's bowl with fan of triangles. E.g., at 1nm resolution,
            # that includes the edge (0, 470).
            for index in range(line_cusp, line_length - 1):
                s = base
                t = s + index
                u = t + 1

                self._faces.append((s, u, t))

                msg = ""
                if trace_enabled:
                    label = "#1" if base == 0 else "#2"
                    msg = f"patch {label}  @ {index:3}"

                for edge in [(s, u), (t, u), (s, t)]:
                    assert edge[0] <= edge[1]
                    edges[edge] += 1

                    if trace_enabled:
                        msg += f" | {edge[0]:7,} -> {edge[1]:7,}"

                trace(msg)

        r = 0
        s = line_length
        t = s - 1
        self._faces.append((r, s, t))

        u = (line_count - 1) * line_length - 1
        v = u + 1
        w = v + line_length - 1
        self._faces.append((u, w, v))

        msg = ""
        if trace_enabled:
            msg = "capstones      "

        for edge in [(r, s), (t, s), (r, t), (u, w), (v, w), (u, v)]:
            assert edge[0] <= edge[1]
            edges[edge] += 1

            if trace_enabled:
                msg += f" | {edge[0]:7,} -> {edge[1]:7,}"

        trace(msg)

        # Check mesh boundary
        boundary = [e for e, count in edges.items() if count == 1]
        if len(boundary) == 0:
            return

        print(
            "\x1b[1;48;5;220mWARNING: Mesh has boundary with "
            f"{len(boundary)} unique edges!\x1b[m"
        )

        if not trace_enabled:
            return

        boundary.sort()
        msg = ""
        for index, edge in enumerate(boundary):
            if index % 6 == 0:
                if index > 0:
                    trace(msg)
                msg = "unique edges   "

            msg += f" | {edge[0]:7,} -> {edge[1]:7,}"

        trace(msg)

    def write_header(
        self,
        file: io.TextIOWrapper,
        *,
        include: None | Self = None,
    ) -> tuple[int, int]:
        file.write("ply\n")
        file.write("format ascii 1.0\n")
        # Blender chokes on PLY files with more than one comment line. Sad!
        file.write(
            f"comment Visual gamut in {self._label}  "
            "<https://github.com/apparebit/prettypretty>\n"
        )

        vertex_count = self.point_count + (0 if include is None else include.point_count)
        file.write(f"element vertex {vertex_count}\n")

        file.write("property float x\n")
        file.write("property float y\n")
        file.write("property float z\n")
        file.write("property uchar red\n")
        file.write("property uchar green\n")
        file.write("property uchar blue\n")
        if self._alpha != 255:
            file.write("property uchar alpha\n")

        face_count = 0
        if self._should_generate_mesh:
            face_count = self.face_count
            file.write(f"element face {face_count}\n")
            file.write("property list uchar int vertex_indices\n")

        file.write("end_header\n")
        return vertex_count, face_count

    def write_vertex_data(self, file: io.TextIOWrapper) -> None:
        for point, color in zip(self._points, self._colors):
            f = 100.0 if self._space is ColorSpace.Xyz else 1.0
            x, y, z = point[0] * f, point[1] * f, point[2] * f
            r, g, b = color
            line = f"{x:f} {y:f} {z:f} {r} {g} {b}"
            if self._alpha < 255:
                line += f" {self._alpha}"
            file.write(f"{line}\n")

    def write_face_data(self, file: io.TextIOWrapper) -> None:
        for i, j, k in self._faces:
            file.write(f"3 {i} {j} {k}\n")

    def write_all(
        self,
        file: io.TextIOWrapper,
        include: None | Self = None,
    ) -> None:
        # Write header
        vertex_count, face_count = self.write_header(file, include=include)

        # Write vertices
        log(f"    {vertex_count:,} vertices")
        self.write_vertex_data(file)
        if include is not None:
            include.write_vertex_data(file)

        # Write faces
        if self._should_generate_mesh:
            log(f"    {face_count:,} faces")
            self.write_face_data(file)

# --------------------------------------------------------- Generate 3D Points and Mesh

def generate(
    *,
    space: ColorSpace,
    filename: str,
    label: str,
    illuminated_observer: IlluminatedObserver,
    stride: int = 2,
    darken: bool = False,
    gamut: None | ColorSpace = None,
    segment_size: int = 200,
    planar_gamut: bool = False,
    mesh: bool = False,
    alpha: float = 1.0,
    sampler: None | Sampler = None,
) -> None:
    log(f"Traversing visual gamut in {BOLD}{label}{RESET}:")
    points = PointManager(
        stride=stride, space=space, darken=darken, mesh=mesh, alpha=alpha, label=label
    )
    is_xyz = space.is_xyz()

    # Lines resulting from square wave pulses
    for step in illuminated_observer.visual_gamut(stride):
        if isinstance(step, GamutTraversalStep.MoveTo):
            points.mark_newline()

        color = step.color()
        if sampler:
            sampler.sample(color)
        if is_xyz:
            points.add(color=color, is_xyz=True)
        else:
            points.add(color=color.to(ColorSpace.Oklrab), is_xyz=False)

    points.generate_faces()

    log(f"    {points.point_count:,} individual points")
    log(f"    {points.line_count:,} lines, each {points.line_length:,} points long")
    log(f"    {points.face_count:,} faces")

    mn = illuminated_observer.minimum()
    mx = illuminated_observer.maximum()
    log(
        f"    min/max in XYZ: {mn[0]:.5f} {mn[1]:.5f} {mn[2]:.5f}   "
        f"{mx[0]:.5f} {mx[1]:.5f} {mx[2]:.5f}"
    )

    gamut_points = None
    if gamut is not None:
        log(f"Traversing {gamut} gamut with segment size {segment_size}")
        gamut_points = PointManager(
            stride=stride, space=space, darken=darken, planar=planar_gamut
        )

        highlight = None

        it = gamut.gamut(segment_size)
        assert it is not None
        for step in it:
            color = step.color()
            if is_xyz:
                gamut_points.add(
                    color=color.to(ColorSpace.Xyz),
                    is_xyz=True,
                    highlight=highlight,
                )
            else:
                gamut_points.add(
                    color=color.to(ColorSpace.Oklrab),
                    is_xyz=False,
                    highlight=highlight,
                )

        log(f"    {gamut_points.point_count:,} points")

    log(f"Writing {BOLD}{filename}{RESET}:")
    with open(filename, mode="w", encoding="utf8") as file:
        points.write_all(file, gamut_points)

# ------------------------------------------------------------------ Interactive Viewer

def render(*, filename: str, label: str, is_ok: bool) -> None:
    try:
        from vedo import ( # pyright: ignore [reportMissingImports, reportMissingTypeStubs]
            Mesh, show # pyright: ignore [reportUnknownVariableType]
        )
    except ImportError:
        log("prettypretty.viz3d requires vedo for rendering the visual gamut.")
        log("Please install the package, e.g., by executing `pip install vedo`,")
        log("and then run `python -m prettypretty.viz3d -mr` again.")
        sys.exit(1)

    mesh: Any = Mesh(filename)
    sx = mesh.clone().project_on_plane('x').c('r').x(0 if is_ok else 0)
    sy = mesh.clone().project_on_plane('y').c('g').y(-0.4 if is_ok else 0)
    sz = mesh.clone().project_on_plane('z').c('b').z(-0.4 if is_ok else 0)

    show(mesh,  # pyright: ignore [reportUnknownMemberType]
        sx, sx.silhouette('2d'), # 2d objects dont need a direction
        sy, sy.silhouette('2d'),
        sz, sz.silhouette('2d'),
        f"The Visual Gamut in {label}",
        axes=7,
        viewup='z',
        bg="#555555",
    ).close() # pyright: ignore [reportOptionalMemberAccess]

# =====================================================================================

if __name__ == "__main__":
    parser = create_parser()
    options = parser.parse_args()

    if options.verbose:
        trace_enabled = True

    # ---------------------------------------------------------------- Colorspace Gamut
    if options.gamut is None:
        gamut = None
    else:
        name = options.gamut.lower()
        if name == "srgb":
            gamut = ColorSpace.Srgb
        elif name == "p3":
            gamut = ColorSpace.DisplayP3
        elif name == "rec2020":
            gamut = ColorSpace.Rec2020
        else:
            log(f"gamut {name} is not srgb, p3, or rec2020 (case-insensitive)!")
            sys.exit(1)

    if options.planar_gamut and gamut is None:
        log("The --planar-gamut option requires the --gamut option")
        sys.exit(1)

    # -------------------------------------------------------------------------- Stride
    stride = 2
    if options.stride:
        try:
            stride = int(options.stride)
        except ValueError:
            log(f"stride {stride} is not an integer")
            sys.exit(1)
    if not (1 <= stride <= 20):
        log(f"stride {stride} is not between 1 and 20 (inclusive)!")
        sys.exit(1)

    # ---------------------------------------------------------------------- Illuminant
    illuminant_name = options.illuminant.upper()
    if illuminant_name == "D50":
        illuminant = CIE_ILLUMINANT_D50
    elif illuminant_name == "D65":
        illuminant = CIE_ILLUMINANT_D65
    elif illuminant_name == "E":
        illuminant = CIE_ILLUMINANT_E
    else:
        raise ValueError(f'invalid value for illuminant "{illuminant_name}"')

    # ------------------------------------------------------------------------ Observer
    if options.observer == "2":
        observer = CIE_OBSERVER_2DEG_1931
    elif options.observer == "10":
        observer = CIE_OBSERVER_10DEG_1964
    else:
        raise ValueError(f'invalid value for observer "{options.observer}"')

    # ---------------------------------------------------------------- Filename & Label
    file_suffix = f"-{stride}nm-{illuminant_name.lower()}-{options.observer}deg.ply"
    label_suffix = f": {stride}nm @ {illuminant_name} * {options.observer}º"

    filename = "visual-gamut/" + ("ok" if options.ok else "xyz") + file_suffix
    label = ("Oklrab" if options.ok else "XYZ") + label_suffix

    # --------------------------------------------------------------------- Render File
    illuminated_observer = IlluminatedObserver(illuminant, observer)
    sampler = Sampler() if options.ok else None
    generate(
        space=ColorSpace.Oklrab if options.ok else ColorSpace.Xyz,
        illuminated_observer=illuminated_observer,
        stride=stride,
        gamut=gamut,
        planar_gamut=options.planar_gamut,
        mesh=options.mesh,
        darken=options.darken,
        alpha=float(options.alpha),
        sampler=sampler,
        filename=filename,
        label=label,
    )

    if options.ok:
        log("\nShape of visible gamut (sampled on chroma/hue plane):\n")
        log(str(sampler))

    # -------------------------------------------------------------- Interactive Viewer
    if options.render:
        render(filename=filename, label=label, is_ok=options.ok)



from collections.abc import Sequence
from typing import ClassVar, Self

# Repeat submodule name to make it public
from . import gamut as gamut
from . import spectrum as spectrum
from . import style as style
from . import termco as termco
from . import theme as theme


class ColorSpace:
    """A color space."""
    Srgb: ClassVar[ColorSpace] = ...
    LinearSrgb: ClassVar[ColorSpace] = ...
    DisplayP3: ClassVar[ColorSpace] = ...
    LinearDisplayP3: ClassVar[ColorSpace] = ...
    Rec2020: ClassVar[ColorSpace] = ...
    LinearRec2020: ClassVar[ColorSpace] = ...
    Oklab: ClassVar[ColorSpace] = ...
    Oklch: ClassVar[ColorSpace] = ...
    Oklrab: ClassVar[ColorSpace] = ...
    Oklrch: ClassVar[ColorSpace] = ...
    Xyz: ClassVar[ColorSpace] = ...
    XyzD50: ClassVar[ColorSpace] = ...

    def is_xyz(self) -> bool: ...
    def is_rgb(self) -> bool: ...
    def is_polar(self) -> bool: ...
    def is_ok(self) -> bool: ...
    def is_bounded(self) -> bool: ...
    def gamut(self, segment_size: int) -> None | gamut.GamutTraversal: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __str__(self) -> str: ...


class OkVersion:
    """The version of the Oklab color space."""
    Original: ClassVar[OkVersion] = ...
    Revised: ClassVar[OkVersion] = ...

    def cartesian_space(self) -> ColorSpace: ...
    def polar_space(self) -> ColorSpace: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...


# ======================================================================================


class Color:
    """High-resolution colors."""
    def __new__(
        cls, space: ColorSpace, coordinates: Sequence[float]
    ) -> Self:
        """
        Create a new color with the color space and coordinates. The coordinates
        must be a sequence with exactly three elements.
        """
        ...
    @staticmethod
    def parse(s: str) -> Color: ...
    @staticmethod
    def srgb(r: float, g: float, b: float) -> Color: ...
    @staticmethod
    def p3(r: float, g: float, b: float) -> Color: ...
    @staticmethod
    def oklab(l: float, a: float, b: float) -> Color: ...
    @staticmethod
    def oklrab(lr: float, a: float, b: float) -> Color: ...
    @staticmethod
    def oklch(l: float, c: float, h: float) -> Color: ...
    @staticmethod
    def oklrch(lr: float, c: float, h: float) -> Color: ...
    @staticmethod
    def from_24bit(r: int, g: int, b: int) -> Color: ...
    def is_default(self) -> bool: ...
    def is_achromatic(self) -> bool: ...
    def is_achromatic_threshold(self, threshold: float) -> bool: ...
    def space(self) -> ColorSpace: ...
    def coordinates(self) -> list[float]: ...
    def normalize(self) -> Self: ...
    def hue_chroma(self) -> tuple[float, float]: ...
    def xy_chromaticity(self) -> tuple[float, float]: ...
    def uv_prime_chromaticity(self) -> tuple[float, float]: ...
    def to(self, target: ColorSpace) -> Self: ...
    def in_gamut(self) -> bool: ...
    def clip(self) -> Self: ...
    def to_gamut(self) -> Self: ...
    def distance(self, other: Self, version: OkVersion) -> float: ...
    def interpolate(
        self,
        color: Self,
        interpolation_space: ColorSpace,
        interpolation_strategy: HueInterpolation,
    ) -> Interpolator: ...
    def lighten(self, factor: float) -> Self: ...
    def darken(self, factor: float) -> Self: ...
    def contrast_against(self, background: Self) -> float: ...
    def use_black_text(self) -> bool: ...
    def use_black_background(self) -> bool: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> float: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def to_24bit(self) -> bytes: ...
    def to_hex_format(self) -> str: ...


class HueInterpolation:
    """The rule for deciding how to interpolate hues."""
    Shorter: ClassVar[HueInterpolation] = ...
    Longer: ClassVar[HueInterpolation] = ...
    Increasing: ClassVar[HueInterpolation] = ...
    Decreasing: ClassVar[HueInterpolation] = ...

    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...


class Interpolator:
    """A color interpolator."""
    def __new__(
        cls,
        color1: Color,
        color2: Color,
        space: ColorSpace,
        strategy: HueInterpolation
    ) -> Self: ...
    def at(self, fraction: float) -> Color: ...


def close_enough(f1: float, f2: float) -> bool: ...


class Translator:
    """A class for translating between terminal and high-resolution colors."""
    def __new__(cls, version: OkVersion, theme: theme.Theme) -> Self: ...
    def __repr__(self) -> str: ...

    # Interrogate the color theme
    def is_dark_theme(self) -> bool: ...

    # Translate terminal to high-resolution colors
    def resolve(
        self,
        color: (
            int | termco.AnsiColor | termco.EmbeddedRgb | termco.GrayGradient |
            termco.EightBitColor | termco.Rgb | Color | termco.Colorant
        ),
    ) -> Color: ...
    def resolve_all(
        self,
        color: (
            int | termco.AnsiColor | termco.EmbeddedRgb | termco.GrayGradient |
            termco.EightBitColor | termco.Rgb | Color | termco.Colorant
        ),
        layer: style.Layer,
    ) -> Color: ...

    # Translate high-resolution to ANSI colors
    def to_ansi(self, color: Color) -> Color: ...
    def supports_hue_lightness(self) -> bool: ...
    def to_ansi_hue_lightness(self, color: Color) -> None | termco.AnsiColor: ...
    def to_closest_ansi(self, color: Color) -> termco.AnsiColor: ...
    def to_ansi_rgb(self, color: Color) -> termco.AnsiColor: ...

    # Translate high-resolution to 8-bit colors
    def to_closest_8bit(self, color: Color) -> termco.EightBitColor: ...
    def to_closest_8bit_with_ansi(self, color: Color) -> termco.EightBitColor: ...

    # Cap terminal colors
    def cap_hires(self, color: Color, fidelity: style.Fidelity) -> None | termco.Colorant: ...
    def cap_colorant(self, color: termco.Colorant, fidelity: style.Fidelity) -> None | termco.Colorant: ...
    def cap(
        self,
        color: (
            int | termco.AnsiColor | termco.EmbeddedRgb | termco.GrayGradient |
            termco.EightBitColor | termco.Rgb | Color | termco.Colorant
        ),
        fidelity: style.Fidelity,
    ) -> None | termco.Colorant: ...



"""
Support for representing the sixteen extended ANSI colors and for mapping them
to meaningful color values.
"""

from collections.abc import Iterator
from contextlib import contextmanager

from .color import OkVersion, Translator
from .color.theme import Theme, VGA_COLORS # pyright: ignore [reportMissingModuleSource]


_current_translator: list[Translator] = [Translator(OkVersion.Revised, VGA_COLORS)]

@contextmanager
def new_theme(theme_colors: Theme) -> Iterator[Translator]:
    """
    Create a new context manager to make the theme colors the current theme
    colors. This function expects exactly 18 colors.
    """
    _current_translator.append(Translator(OkVersion.Revised, theme_colors))
    try:
        yield _current_translator[-1]
    finally:
        _current_translator.pop()

def current_translator() -> Translator:
    """
    Access the current translator.

    The translator is automatically updated when using different theme colors.
    """
    return _current_translator[-1]



from typing import ClassVar, Self

from . import Color, Translator
from .termco import AnsiColor, Colorant, EightBitColor, EmbeddedRgb, GrayGradient, Rgb


class Attribute:
    """A text formatting attribute."""
    Bold: ClassVar[Attribute] = ...
    Thin: ClassVar[Attribute] = ...
    Italic: ClassVar[Attribute] = ...
    Underlined: ClassVar[Attribute] = ...
    Blinking: ClassVar[Attribute] = ...
    Reversed: ClassVar[Attribute] = ...
    Hidden: ClassVar[Attribute] = ...
    Stricken: ClassVar[Attribute] = ...

    def enable_sgr(self) -> int: ...
    def disable_sgr(self) -> int: ...
    def __add__(self, other: Self) -> FormatUpdate: ...
    def __sub__(self, other: Self) -> FormatUpdate: ...
    def __neg__(self) -> FormatUpdate: ...
    def __repr__(self) -> str: ...


class Format:
    """A text format comprising zero or more attributes."""
    @staticmethod
    def of(formatting: Attribute | Format) -> Format: ...

    def __new__(cls) -> Self: ...
    def is_empty(self) -> bool: ...
    def len(self) -> int: ...
    def __len__(self) -> int: ...
    def attributes(self) -> AttributeIter: ...
    def __add__(self, other: Attribute | Self) -> FormatUpdate: ...
    def __sub__(self, other: Attribute | Self) -> FormatUpdate: ...
    def __neg__(self) -> FormatUpdate: ...
    def __repr__(self) -> str: ...


class AttributeIter:
    """An iterator over format attributes."""
    def __len__(self) -> int: ...
    def __iter__(self) -> Self: ...
    def __next__(self) -> Attribute: ...


class FormatUpdate:
    """
    A text format update comprising a format to disable and a format to enable.
    """
    @staticmethod
    def of(formatting: Attribute | Format | FormatUpdate) -> FormatUpdate: ...

    def __new__(cls) -> Self: ...
    def is_empty(self) -> bool: ...
    def disable(self) -> Format: ...
    def enable(self) -> Format: ...
    def cap(self, fidelity: Fidelity) -> Self: ...
    def __add__(self, other: Attribute | Self) -> FormatUpdate: ...
    def __sub__(self, other: Attribute | Self) -> FormatUpdate: ...
    def __neg__(self) -> FormatUpdate: ...
    def __repr__(self) -> str: ...


class Fidelity:
    """
    Levels of support for ANSI escape codes and colors.
    """
    Plain: ClassVar[Fidelity] = ...
    NoColor: ClassVar[Fidelity] = ...
    Ansi: ClassVar[Fidelity] = ...
    EightBit: ClassVar[Fidelity] = ...
    TwentyFourBit: ClassVar[Fidelity] = ...
    HiRes: ClassVar[Fidelity] = ...

    @staticmethod
    def from_color(
        color: (
            int | AnsiColor | EmbeddedRgb | GrayGradient | EightBitColor | Rgb
            | Color | Colorant
        )
    ) -> Fidelity: ...
    @staticmethod
    def from_environment(has_tty: bool) -> Fidelity: ...

    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __le__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...


class Layer:
    """The render layer."""
    Foreground: ClassVar[Layer] = ...
    Background: ClassVar[Layer] = ...

    def is_foreground(self) -> bool: ...
    def is_background(self) -> bool: ...
    def offset(self) -> int: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...


class Style:
    """A style."""
    # Methods to configure styles
    def __new__(cls) -> Self: ...
    def bold(self) -> Self: ...
    def thin(self) -> Self: ...
    def italic(self) -> Self: ...
    def underlined(self) -> Self: ...
    def blinking(self) -> Self: ...
    def reversed(self) -> Self: ...
    def hidden(self) -> Self: ...
    def stricken(self) -> Self: ...
    def with_foreground(
        self,
        color: (
            int | AnsiColor | EmbeddedRgb | GrayGradient | EightBitColor | Rgb
            | Color | Colorant
        ),
    ) -> Self: ...
    def with_background(
        self,
        color: (
            int | AnsiColor | EmbeddedRgb | GrayGradient | EightBitColor | Rgb
            | Color | Colorant
        ),
    ) -> Self: ...

    # Method to inspect styles
    def is_default(self) -> bool: ...
    def format(self) -> Format: ...
    def foreground(self) -> None | Colorant: ...
    def background(self) -> None | Colorant: ...
    def fidelity(self) -> Fidelity: ...
    def cap(self, fidelity: Fidelity, translator: Translator) -> Self: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __neg__(self) -> Self: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
